/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// HexRecord objects store hexadecimal records consisting of
// a length, beginning address, hexadecimal data values, checksum,
// and tag.  These records are usually created by reading in
// an Intel hex file.
//

// $Log: HEXRECRD.CPP,v $
// Revision 1.2  2000/11/14 11:25:15  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.3  1999/07/06 09:06:33  DEVB
// changed printing of hex records to upper-case strings.
//
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 21:52:36  DEVB
// Initial revision
//

#include <assert.h>
#include <stdlib.h>
#include <iomanip.h>
#include <ctype.h>

#include "HEX.H"
#include "UTILS.H"
#include "HEXRECRD.H"



// begin methods and functions

// construct and initialize an empty hex record
HexRecord::HexRecord(void)
{
	data = NULL;				// no hex data at this time
	SetTag(InvalidRecord);		// record invalid until it's loaded with something
	SetLength(0);				// length of hex data is zero
	SetAddress(0);				// hex data is not destined for any address
	SetError(NoHexRecordError);	// no errors yet
	CalcCheckSum();				// calculate the checksum for this trivial record
}

// delete a hex record
HexRecord::~HexRecord(void)
{
	delete data;
}

// set the type tag of the hex record
void HexRecord::SetTag(HexRecordTagType t)
{
	tag = t;
}

// get the tag of the hex record
HexRecordTagType HexRecord::GetTag(void) const
{
	return tag;
}

// does the hex record contain data?
bool HexRecord::IsData(void) const
{
	return tag==DataRecord || tag==NoSumRecord;
}

// is the hex record valid?
bool HexRecord::IsValid(void) const
{
	return tag!=InvalidRecord;
}

// Set the size of the data area in the hex record.  Delete existing data.
void HexRecord::SetLength(unsigned int l)
{
	if(length != 0)
		delete data;
	length = l;
	data = new unsigned char[length];
}

// get the number of data bytes in the hex record
unsigned int HexRecord::GetLength(void) const
{
	return length;
}

// set the address at which the hex record data should be loaded
void HexRecord::SetAddress(unsigned int addr)
{
	address = addr;
}

// get address at which hex record data will be loaded
unsigned int HexRecord::GetAddress(void) const
{
	return address;
}

// index into the hex record to get a single byte
unsigned char& HexRecord::operator[](unsigned int index) const
{
	assert(index<length);
	return data[index];
}

// calculate the 8-bit checksum of the data in the hex record
void HexRecord::CalcCheckSum(void)
{
	int i;
	
	if(tag != NoSumRecord)
	{ // hex record requires a checksum
		for(checkSum=0, i=length-1; i>=0; i--)
			checkSum += data[i];
		checkSum += length;
		checkSum += tag;
		checkSum += address & 0xFF;
		checkSum += ((address>>8) & 0xFF);
		checkSum = (-checkSum) & 0xFF;
	}
	else	// hex record requires no checksum
		checkSum = 0;	// default checksum value for
}

// get the calculated checksum of the hex record
unsigned char HexRecord::GetCheckSum(void) const
{
	return checkSum;
}

// set the error flag in the hex record
void HexRecord::SetError(HexRecordError e)
{
	err = e;
}

// get the error flag from the hex record
HexRecordError HexRecord::GetError(void) const
{
	return err;
}

// is there an error in the hex record
bool HexRecord::IsError(void) const
{
	return err != NoHexRecordError;
}

// return an error message for the hex record
const char* HexRecord::GetErrMsg(void) const
{
	return ErrMsg(GetError());
}

// return an error message for the given type of error
const char* ErrMsg(HexRecordError e)
{
	switch(e)
	{
	case NoHexRecordError:
		return "no error";
	case StartHexRecordError:
		return "should start with \':\'";
	case LengthHexRecordError:
		return "invalid length";
	case AddressHexRecordError:
		return "invalid address";
	case TagHexRecordError:
		return "invalid tag";
	case DataHexRecordError:
		return "invalid hex data";
	case CheckSumHexRecordError:
		return "invalid checksum";
	default:
		return "unknown error";
	}
}


// helper function to load a hex record from a stream
istream& operator>> (istream& is, HexRecord& hx)
{
	unsigned char c;	// character from input stream
	Hex length(2);		// two-digit length for hex data record
	Hex tag(2);			// two-digit data record tag
	Hex address(4);		// four-digit destination address for hex data
	Hex data(2);		// two-digit data item
	Hex checkSum(2);	// two-digit checksum
	
	// get the initial character of a hex record to see what type it is
	is >> c;
	if(is.eof())
	{ // no hex record error
		hx.SetError(NoHexRecordError);
		hx.SetTag(InvalidRecord);
		return is;
	}
	if(is.fail())
	{ // error starting hex record
		hx.SetError(StartHexRecordError);
		hx.SetTag(InvalidRecord);
		return is;
	}
	if(c!=':')
	{ // hex records that do not start with ':' also do not have checksums
		hx.SetTag(NoSumRecord);
	}
	
	// now get the number of bytes in the hex record
	is >> length;
	if(is.fail())
	{ // no length given for the hex record
		hx.SetError(LengthHexRecordError);
		hx.SetTag(InvalidRecord);
		return is;
	}
	hx.SetLength(length.GetHex());
	
	// now get the destination address for the data bytes
	is >> address;
	if(is.fail())
	{ // no starting address for the hex record data
		hx.SetError(AddressHexRecordError);
		hx.SetTag(InvalidRecord);
		return is;
	}
	hx.SetAddress(address.GetHex());
	
	// set type of hex record if it's not already set to a noSumRecord
	if(hx.GetTag()!=NoSumRecord)
	{ // this is the standard Intel hex record format, so look for the type
		is >> tag;
		if(is.fail())
		{ // no record type error
			hx.SetError(TagHexRecordError);
			hx.SetTag(InvalidRecord);
			return is;
		}
		hx.SetTag((HexRecordTagType)(tag.GetHex())); // record the type of record
	}
	switch(hx.GetTag())
	{ // check the record type to see if it's valid
	case DataRecord:
	case InfoRecord:
	case Info2Record:
	case NoSumRecord:
		break; // all the tag types given above are ok
	default:
		hx.SetError(TagHexRecordError);
		hx.SetTag(InvalidRecord);
		return is;
	}
	
	// read in the number of data bytes given in the length
    for(unsigned int i=0; i<hx.GetLength(); i++)
	{
		is >> data;
		hx[i] = data.GetHex();
		if(is.fail())
		{ // error if data ended too soon
			hx.SetError(DataHexRecordError);
			hx.SetTag(InvalidRecord);
			return is;
		}
	}
	
	// calculate the checksum for the data that has been read
	if(hx.GetTag()!=NoSumRecord)
	{ // if this is a hex record with a checksum, compare it with
		// the calculated value
		is >> checkSum; // get the checksum for the hex record
		if(is.fail())
		{ // error if no checksum when there should be one
			hx.SetError(CheckSumHexRecordError);
			hx.SetTag(InvalidRecord);
			return is;
		}
		hx.CalcCheckSum();	// now calculate the checksum for the data
		if(checkSum.GetHex()!=hx.GetCheckSum())
		{ // error if checksums do not match
			hx.SetError(CheckSumHexRecordError);
			hx.SetTag(InvalidRecord);
			return is;
		}
	}
	
	// got all the way to the end, so there are no errors in the hex record
	hx.SetError(NoHexRecordError);
	return is;
}

// helper function to output the hex record to a stream in Intel format
ostream& operator<< (ostream& os, HexRecord& hx)
{
	if(!hx.IsValid())
		return os;	// abort if hex record is not valid
	
	os << ":";
	os << setiosflags(ios::uppercase) << setfill('0') << setw(2) << hex << hx.GetLength();
	os << setiosflags(ios::uppercase) << setfill('0') << setw(4) << hex << hx.GetAddress();
	os << setiosflags(ios::uppercase) << setfill('0') << setw(2) << hex << hx.GetTag();
	for(unsigned int i=0; i<hx.GetLength(); i++)
		os << setiosflags(ios::uppercase) << setfill('0') << setw(2) << hex << (int)hx[i];
	os << setiosflags(ios::uppercase) << setfill('0') << setw(2) << hex << (int)hx.GetCheckSum();
	os << endl;
	return os;
}
