/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// HexRecord objects store hexadecimal records consisting of
// a length, beginning address, hexadecimal data values, checksum,
// and tag.  These records are usually created by reading in
// an Intel hex file.
//

// $Log: HEXRECRD.H,v $
// Revision 1.2  2000/11/14 11:25:15  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.2  1999/03/30 18:27:41  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:52:12  DEVB
// Initial revision
//

#ifndef HexRecord_H
#define HexRecord_H

#include <iostream.h>

// these are the various types of records found in a hex file
typedef enum
{
	DataRecord=0,		// actual address and data
		InfoRecord=1,		// some type of informational record
		Info2Record=2,		// another type of informational record
		InvalidRecord=3,	// erroneous record
		NoSumRecord=4,		// for storing bytes with no checksum
} HexRecordTagType;

// various types of errors that can occur when handling hex records
typedef enum
{
	NoHexRecordError,
		StartHexRecordError,
		LengthHexRecordError,
		AddressHexRecordError,
		TagHexRecordError,
		DataHexRecordError,
		CheckSumHexRecordError
} HexRecordError;


class HexRecord
{
	public:

	/* begin HexRecord methods */

	/* method: HexRecord */
	// construct and initialize an empty hex record
	 HexRecord(void);

	/* method: ~HexRecord */
	// delete a hex record
	 ~HexRecord(void);

	/* method: SetTag */
	// set the type tag of the hex record
	void SetTag(HexRecordTagType t);

	/* method: GetTag */
	// get the tag of the hex record
	HexRecordTagType GetTag(void) const;

	/* method: IsData */
	// does the hex record contain data?
	bool IsData(void) const;

	/* method: IsValid */
	// is the hex record valid?
	bool IsValid(void) const;

	/* method: SetLength */
	// Set the size of the data area in the hex record.  Delete existing data.
	void SetLength(unsigned int l);

	/* method: GetLength */
	// get the number of data bytes in the hex record
	unsigned int GetLength(void) const;

	/* method: SetAddress */
	// set the address at which the hex record data should be loaded
	void SetAddress(unsigned int addr);

	/* method: GetAddress */
	// get address at which hex record data will be loaded
	unsigned int GetAddress(void) const;

	/* method: operator[] */
	// index into the hex record to get a single byte
	unsigned char& operator[](unsigned int index) const;

	/* method: CalcCheckSum */
	// calculate the 8-bit checksum of the data in the hex record
	void CalcCheckSum(void);

	/* method: GetCheckSum */
	// get the calculated checksum of the hex record
	unsigned char GetCheckSum(void) const;

	/* method: SetError */
	// set the error flag in the hex record
	void SetError(HexRecordError e);

	/* method: GetError */
	// get the error flag from the hex record
	HexRecordError GetError(void) const;

	/* method: IsError */
	// is there an error in the hex record
	bool IsError(void) const;

	/* method: GetErrMsg */
	// return an error message for the hex record
	const char* GetErrMsg(void) const;

	private:

	HexRecordError err;			// hex record error
	HexRecordTagType tag;		// type of data stored in hex record
	unsigned int length;		// number of data bytes in the hex record
	unsigned int address;		// destination address for the hex data
	unsigned char* data;		// data values stored in the hex record
	unsigned char checkSum;		// checksum for the entire hex record

	/* end HexRecord methods */

};

/* begin HexRecord functions */

/* function: ErrMsg */
// return an error message for the given type of error
const char* ErrMsg(HexRecordError e);

// helper function that reads in a hex record from an input stream
extern istream& operator>> (istream& is, HexRecord& hx);

// helper function that dumps a hex record to an output stream
extern ostream& operator<< (ostream& is, HexRecord& hx);

/* end HexRecord functions */

#endif
