/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// Pport objects handle the parallel port data, status, and
// control pins.
//

// $Log: PPORT.H,v $
// Revision 1.2  2000/11/14 11:25:16  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.2  1999/03/30 18:27:41  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:53:42  DEVB
// Initial revision
//

#ifndef Pport_H
#define Pport_H

#include "XSERROR.H"

class Pport
{
	public:

	/* begin Pport methods */

	/* method: Pport */
	// control register address offset
	// Constructor for a parallel port object.  Inversion masks are used to
	// correct the effects of the various inverters in the PC and XS Board.
	 Pport(XSError* e, // error reporting channel 
		unsigned int n=1, // parallel port number
		unsigned int dInvMask=0, // inversion mask for data output bits
		unsigned int sInvMask=0, // inversion mask for status input bits
		unsigned int cInvMask=0 // inversion mask for control output bits
		);

	/* method: ~Pport */
	// Destructor for parallel port object
	 ~Pport(void);

	/* method: GetErr */
	// provides access to the error reporting channel
	XSError& GetErr(void);

	/* method: operator= */
	// Assignment operator for parallel port objects
	Pport& operator=(Pport& src);

	/* method: SetNum */
	// Sets up a port object for a given parallel port n
	void SetNum(unsigned int n);

	/* method: GetNum */
	// Gets the parallel port number for a parallel port object
	unsigned int GetNum(void) const;

	// 	Determines I/O address for the given parallel port # (1-4).
	void FindIOAddress(void);

	/* method: SetDataInvMask */
	// Set the data register inversion mask
	void SetDataInvMask(unsigned int invMask=0);

	/* method: SetData */
	// Output a value on the designated data pins of the parallel port
	void SetData(
		unsigned int v, // value to output
		unsigned int loPos=0, // low bit position of field
		unsigned int hiPos=7 // high bit position of field
		);

	/* method: GetData */
	// Return the current values on the designated data pins of the parallel port
	unsigned int GetData(
		unsigned int loPos=0, // low bit position of field
		unsigned int hiPos=7 // high bit position of field
		) const;

	/* method: SetStatusInvMask */
	// Set the status register inversion mask
	void SetStatusInvMask(unsigned int invMask=0);

	/* method: GetStatus */
	// Return the current values on the designated status pins of the parallel port
	unsigned int GetStatus(
		unsigned int loPos=0, // low bit position of field
		unsigned int hiPos=7 // high bit position of field
		) const;

	/* method: SetControlInvMask */
	// Set the control register inversion mask
	void SetControlInvMask(unsigned int invMask=0);

	/* method: SetControl */
	// Output a value on the designated control pins of the parallel port
	void SetControl(
		unsigned int v, // value to output
		unsigned int loPos=0, // low bit position of field
		unsigned int hiPos=7 // high bit position of field
		);

	/* method: GetControl */
	// Return the current values on the designated control pins of the parallel port
	unsigned int GetControl(
		unsigned int loPos=0, // low bit position of field
		unsigned int hiPos=7 // high bit position of field
		) const;

	/* end Pport methods */

	private:

	XSError* err;					// error reporting object
	unsigned int num;				// parallel port num 1,2,3,4
        int lp0;    // handle to parellel port0
        unsigned int address;			// I/O address
	unsigned int dataInvMask;		// inversion mask for data outputs
	unsigned int statusInvMask;		// inversion mask for status inputs
	unsigned int controlInvMask;	// inversion mask for control outputs
};

/* begin Pport functions */

/* function: Inp */
//
// Here are some routines for handling primitive read/write of ports
//
// get data byte from a given 16-bit I/O address
static unsigned char Inp(unsigned short address);

/* function: Outp */
// send data byte to a given 16-bit I/O address
static void Outp(unsigned short address, unsigned char byte);

/* function: CheckBitPos */
// Check for bit position index which is too large
static bool CheckBitPos(unsigned int pos, // bit position
	XSError& err // error reporting channel
	);

/* function: GetField */
// Get the value stored in a bit field
static unsigned int GetField(unsigned int data, // port data
	unsigned int loPos, // low bit position of field
	unsigned int hiPos // high bit position of field
	);

/* function: SetField */
// Set the value in a bit field and return the entire data value after the
// field is updated.
static unsigned int SetField(unsigned int data, // port data
	unsigned int loPos, // low bit position of field
	unsigned int hiPos, // high bit position of field
	unsigned int newData // new data for field
	);

/* end Pport functions */

#endif
