/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// S-record objects store hexadecimal records consisting of
// a length, beginning address, hexadecimal data values, and checksum.
// These records are usually created by reading in
// a Motorola Exormacs file.
//

// $Log: SRECRD.CPP,v $
// Revision 1.3  2000/12/11 15:29:16  bretzdl
// exo file download support added, asm inline removed
//
// Revision 1.2  2000/11/14 11:25:16  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.1  2000/09/10 15:53:35  DEVB
// Initial revision
//

#include <assert.h>
#include <stdlib.h>
#include <iomanip.h>
#include <ctype.h>

#include "HEX.H"
#include "UTILS.H"
#include "SRECRD.H"



// begin methods and functions

// construct and initialize an empty S-record
SRecord::SRecord(void)
{
	data = NULL;				// no hex data at this time
	SetTag(InvalidSRecord);		// record invalid until it's loaded with something
	SetLength(0);				// length of hex data is zero
	SetAddress(0);				// hex data is not destined for any address
	SetError(NoSRecordError);	// no errors yet
	CalcCheckSum();				// calculate the checksum for this trivial record
}

// delete an S-record
SRecord::~SRecord(void)
{
	delete data;
}

// set the type tag of the S-record
void SRecord::SetTag(SRecordTagType t)
{
	tag = t;
}

// get the tag of the S-record
SRecordTagType SRecord::GetTag(void) const
{
	return tag;
}

// does the S-record contain data?
bool SRecord::IsData(void) const
{
	return tag==DataSRecord;
}

// is the S-record valid?
bool SRecord::IsValid(void) const
{
	return tag!=InvalidSRecord;
}

// Set the size of the data area in the S-record.  Delete existing data.
void SRecord::SetLength(unsigned int l)
{
	if(length != 0)
		delete data;
	length = l;
	data = new unsigned char[length];
}

// get the number of data bytes in the S-record
unsigned int SRecord::GetLength(void) const
{
	return length;
}

// set the address at which the S-record data should be loaded
void SRecord::SetAddress(unsigned int addr)
{
	address = addr;
}

// get address at which S-record data will be loaded
unsigned int SRecord::GetAddress(void) const
{
	return address;
}

// set the number of bytes in the address at which the S-record data should be loaded
void SRecord::SetNumAddressBytes(unsigned int numBytes)
{
	numAddressBytes = numBytes;
}

// get the number of bytes in the address at which S-record data will be loaded
unsigned int SRecord::GetNumAddressBytes(void) const
{
	return numAddressBytes;
}

// index into the S-record to get a single byte
unsigned char& SRecord::operator[](unsigned int index) const
{
	assert(index<length);
	return data[index];
}

// calculate the 8-bit checksum of the data in the S-record
void SRecord::CalcCheckSum(void)
{
	int i;
	for(checkSum=0, i=length-1; i>=0; i--)
		checkSum += data[i];
	// add # data bytes, address bytes, and checksum byte to checksum
	checkSum += (length + numAddressBytes+1);
	unsigned int addr = address;
	for(i=numAddressBytes; i>0; i--)
	{
		checkSum += addr & 0xFF;
		addr >>= 8;
	}
	checkSum = (0xFF - checkSum) & 0xFF;
}

// get the calculated checksum of the S-record
unsigned char SRecord::GetCheckSum(void) const
{
	return checkSum;
}

// set the error flag in the S-record
void SRecord::SetError(SRecordError e)
{
	err = e;
}

// get the error flag from the S-record
SRecordError SRecord::GetError(void) const
{
	return err;
}

// is there an error in the S-record
bool SRecord::IsError(void) const
{
	return err != NoSRecordError;
}

// return an error message for the S-record
const char* SRecord::GetErrMsg(void) const
{
	return ErrMsg(GetError());
}

// return an error message for the given type of error
const char* ErrMsg(SRecordError e)
{
	switch(e)
	{
	case NoSRecordError:
		return "no error";
	case StartSRecordError:
		return "should start with \'S\'";
	case LengthSRecordError:
		return "invalid length";
	case AddressSRecordError:
		return "invalid address";
	case TagSRecordError:
		return "invalid tag";
	case DataSRecordError:
		return "invalid hex data";
	case CheckSumSRecordError:
		return "invalid checksum";
	default:
		return "unknown error";
	}
}


// helper function to load an S-record from a stream
istream& operator>> (istream& is, SRecord& sr)
{
	unsigned char c;	// character from input stream
	Hex length(2);		// two-digit length for hex data in S-record
	Hex tag(2);			// two-digit data record tag
	Hex address(6);		// six-digit destination address for hex data
	int numAddressBytes;	// number of bytes in destination address
	Hex data(2);		// two-digit data item
	Hex checkSum(2);	// two-digit checksum
	
	// get the initial character of an S-record to see what type it is
	is >> c;
	if(is.eof())
	{ // no hex record error
		sr.SetError(NoSRecordError);
		sr.SetTag(InvalidSRecord);
		return is;
	}
	if(is.fail())
	{ // error starting S-record
		sr.SetError(StartSRecordError);
		sr.SetTag(InvalidSRecord);
		return is;
	}
	if(c!='S')
	{ // S-records that do no start with 'S' are in error
		sr.SetError(StartSRecordError);
		sr.SetTag(InvalidSRecord);
		return is;
	}
	is >> c;
	if(is.eof())
	{ // no hex record error
		sr.SetError(NoSRecordError);
		sr.SetTag(InvalidSRecord);
		return is;
	}
	if(is.fail())
	{ // error starting S-record
		sr.SetError(StartSRecordError);
		sr.SetTag(InvalidSRecord);
		return is;
	}
	
	numAddressBytes = 2;	// smallest address size (2 bytes)
	switch(c)
	{
	case '3':	// data record with 4-byte address
		numAddressBytes++;
	case '2':	// data record with 3-byte address
		numAddressBytes++;
	case '1':	// data record with 2-byte address
		sr.SetTag(DataSRecord);
		break;

	case '7':	// starting execution address in 4-byte address
		numAddressBytes++;
	case '8':	// starting execution address in 3-byte address
		numAddressBytes++;
	case '9':	// starting execution address in 2-byte address
	case '5':	// # of S1,S2,S3 records so far stored in 2-byte address
	case '0':	// name, version, revision info record with 2-byte address
		sr.SetTag(InfoSRecord);
		break;

	default:
		sr.SetError(StartSRecordError);
		sr.SetTag(InvalidSRecord);
		return is;
	}
		
	// now get the number of bytes in the S-record
	is >> length;
	if(is.fail())
	{ // no length given for the S-record
		sr.SetError(LengthSRecordError);
		sr.SetTag(InvalidSRecord);
		return is;
	}
	// reduce the length by the number of address bytes and the checksum byte
	sr.SetLength(length.GetHex()-numAddressBytes-1);
	
	// now get the destination address for the data bytes
	address.SetLength(2 * numAddressBytes);	// set # of hex digits in address
	is >> address;
	if(is.fail())
	{ // no starting address for the S-record data
		sr.SetError(AddressSRecordError);
		sr.SetTag(InvalidSRecord);
		return is;
	}
	sr.SetAddress(address.GetHex());
	sr.SetNumAddressBytes(numAddressBytes);
	
	switch(sr.GetTag())
	{ // check the record type to see if it's valid
	case DataSRecord:
	case InfoSRecord:
		break; // all the tag types given above are ok
	default:
		sr.SetError(TagSRecordError);
		sr.SetTag(InvalidSRecord);
		return is;
	}
	
	// read in the number of data bytes given in the length
    for(unsigned int i=0; i<sr.GetLength(); i++)
	{
		is >> data;
		sr[i] = data.GetHex();
		if(is.fail())
		{ // error if data ended too soon
			sr.SetError(DataSRecordError);
			sr.SetTag(InvalidSRecord);
			return is;
		}
	}
	
	// calculate the checksum for the data that has been read
	is >> checkSum; // get the checksum for the S-record
	if(is.fail())
	{ // error if no checksum when there should be one
		sr.SetError(CheckSumSRecordError);
		sr.SetTag(InvalidSRecord);
		return is;
	}
	sr.CalcCheckSum();	// now calculate the checksum for the data
	if(checkSum.GetHex()!=sr.GetCheckSum())
	{ // error if checksums do not match
		sr.SetError(CheckSumSRecordError);
		sr.SetTag(InvalidSRecord);
		return is;
	}
	
	// got all the way to the end, so there are no errors in the S-record
	sr.SetError(NoSRecordError);
	return is;
}

// helper function to output the S-record to a stream in Motorola Xormacs format
ostream& operator<< (ostream& os, SRecord& sr)
{
	if(!sr.IsValid())
		return os;	// abort if S-record is not valid
	
	os << (sr.GetTag()==DataSRecord ? "S2" : "S8");
	os << setiosflags(ios::uppercase) << setfill('0') << setw(2) << hex << sr.GetLength()+4;
	os << setiosflags(ios::uppercase) << setfill('0') << setw(6) << hex << sr.GetAddress();
	for(unsigned int i=0; i<sr.GetLength(); i++)
		os << setiosflags(ios::uppercase) << setfill('0') << setw(2) << hex << (int)sr[i];
	os << setiosflags(ios::uppercase) << setfill('0') << setw(2) << hex << (int)sr.GetCheckSum();
	os << endl;
	return os;
}
