/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// S-record objects store hexadecimal records consisting of
// a length, beginning address, hexadecimal data values, checksum,
// and tag.  These records are usually created by reading in
// a Motoroal Exormacs file.
//

// $Log: SRECRD.H,v $
// Revision 1.2  2000/11/14 11:25:16  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.1  2000/09/10 15:53:35  DEVB
// Initial revision
//

#ifndef SRecord_H
#define SRecord_H

#include <iostream.h>

// these are the various types of records found in an S-record file
typedef enum
{
	DataSRecord=0,		// actual address and data
	InfoSRecord=1,		// some type of informational record
	InvalidSRecord=3,	// erroneous record
} SRecordTagType;

// various types of errors that can occur when handling hex records
typedef enum
{
	NoSRecordError,
		StartSRecordError,
		LengthSRecordError,
		AddressSRecordError,
		TagSRecordError,
		DataSRecordError,
		CheckSumSRecordError
} SRecordError;


class SRecord
{
	public:

	/* begin SRecord methods */

	/* method: SRecord */
	// construct and initialize an empty S-record
	 SRecord(void);

	/* method: ~SRecord */
	// delete an S-record
	 ~SRecord(void);

	/* method: SetTag */
	// set the type tag of the S-record
	void SetTag(SRecordTagType t);

	/* method: GetTag */
	// get the tag of the S-record
	SRecordTagType GetTag(void) const;

	/* method: IsData */
	// does the S-record contain data?
	bool IsData(void) const;

	/* method: IsValid */
	// is the S-record valid?
	bool IsValid(void) const;

	/* method: SetLength */
	// Set the size of the data area in the S-record.  Delete existing data.
	void SetLength(unsigned int l);

	/* method: GetLength */
	// get the number of data bytes in the S-record
	unsigned int GetLength(void) const;

	/* method: SetAddress */
	// set the address at which the S-record data should be loaded
	void SetAddress(unsigned int addr);

	/* method: GetAddress */
	// get address at which S-record data will be loaded
	unsigned int GetAddress(void) const;

	/* method: SetNumAddressBytes */
	// set the number of bytes in the address at which the S-record data should be loaded
	void SetNumAddressBytes(unsigned int numBytes);

	/* method: GetNumAddressBytes */
	// get the number of bytes in the address at which S-record data will be loaded
	unsigned int GetNumAddressBytes(void) const;

	/* method: operator[] */
	// index into the S-record to get a single byte
	unsigned char& operator[](unsigned int index) const;

	/* method: CalcCheckSum */
	// calculate the 8-bit checksum of the data in the S-record
	void CalcCheckSum(void);

	/* method: GetCheckSum */
	// get the calculated checksum of the S-record
	unsigned char GetCheckSum(void) const;

	/* method: SetError */
	// set the error flag in the S-record
	void SetError(SRecordError e);

	/* method: GetError */
	// get the error flag from the S-record
	SRecordError GetError(void) const;

	/* method: IsError */
	// is there an error in the S-record
	bool IsError(void) const;

	/* method: GetErrMsg */
	// return an error message for the S-record
	const char* GetErrMsg(void) const;

	private:

	SRecordError err;			// S-record error
	SRecordTagType tag;			// type of data stored in S-record
	unsigned int length;		// number of data bytes in the S-record
	unsigned int address;		// destination address for the S-record data
	unsigned int numAddressBytes;	// number of bytes in destination address
	unsigned char* data;		// data values stored in the S-record
	unsigned char checkSum;		// checksum for the entire S-record

	/* end SRecord methods */

};

/* begin SRecord functions */

/* function: ErrMsg */
// return an error message for the given type of error
const char* ErrMsg(SRecordError e);

// helper function that reads in an S-record from an input stream
extern istream& operator>> (istream& is, SRecord& hx);

// helper function that dumps an S-record to an output stream
extern ostream& operator<< (ostream& is, SRecord& hx);

/* end SRecord functions */

#endif
