/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The XC40 object specializes the XCPORT object so that the XC4000 FPGA on the
// XS40 Board can be configured and the RAM can be loaded.  The XC40 object controls
// the the program initiation, clock, and data pins.
//

// $Log: XC40.CPP,v $
// Revision 1.2  2000/11/14 11:25:16  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.9  2000/09/10 13:33:56  DEVB
// Replaced endl with newline to avoid problems with logging
// of error messages by the xserror object.
//
// Revision 1.8  1999/12/10 02:44:13  DEVB
// Changed Serial EEPROM programming algorithm so it also sets the
// polarity of the Atmel EEPROM RESET/OE so the OE is active-high and
// the reset is active-low.
//
// Revision 1.7  1999/09/26 19:18:59  DEVB
// added short delay after configuration to allow the chip to initialize.
//
// Revision 1.6  1999/09/24 18:04:24  DEVB
// added routine to upload XS40 RAM to PC.
// changed download routines so they no longer use the pseudo-JTAG interface.
//
// Revision 1.5  1999/07/06 09:07:57  DEVB
// changed names of routines that configure FPGA for downloading to the RAM
// to reflect the fact that these same routines are now used to configure
// the FPGA for uploading from the RAM.
//
// Revision 1.4  1999/06/29 16:54:30  DEVB
// added two more address bits to support 128 KByte RAMs.
//
// Revision 1.3  1999/06/22 18:54:24  DEVB
// made changes to allow access to pins which control the EEPROM and
// programmable oscillator.
//
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:00:48  DEVB
// Initial revision
//

#include <fstream.h>
#include "XCBSDR.H"
#include "XC40.H"

// begin methods and functions

// Create an XC40 object.  The type of XC4000 FPGA on the XS40 Board is inferred
// by reading the chip type from a bitfile that is passed to the constructor.
// The directory containing the RAM interface bitstreams for all the different types
// of XC4000 chips is also passed as an argument.
XC40::XC40(XSError* e,	// error reporting channel
		   unsigned int portNum,	// parallel port number
		   string& bitfileName,		// bitfile with XC4000 configuration bitstream
		   string& intfcDir			// directory where RAM interface bitstreams are stored
		   ) : XCPort(e,portNum)	// instantiate the XCPort base object
{	
	// make sure the PROGRAM pin is high
	SetPROGRAM(1);
	// make sure configuration clock is low
	SetCCLK(0);
	// Pull all the data pins high.  The data pins control FPGA circuitry that controls other
	// chips.  The LSB controls the active-low chip-enable of the serial EEPROM.
	// Bits 2 and 3 control the clock pin of the programmable oscillator which is used to set the
	// frequency.  Pulling bit 3 high tristates the FPGA driver to the clock pin so the oscillator can
	// output a signal without interference.
	SetData(0xff);
	// get the chip identifier from the bitstream file
	SetChipType(bitfileName);
	// store the directory where the RAM interface bitstreams are stored
	ramInterfaceDir = intfcDir;
	ramInterfaceIsRdy = false;	// chip has not yet been loaded with the RAM interface bitstream
}


// These are the bit positions of the FPLD configuration control lines
// for the parallel port interface of the XS40 board.
static const unsigned int posPROG = 0;	// output bit to initiate configuration input pin of XS Board
static const unsigned int posCCLK = 1;	// output bit to configuration clock input of XS board
static const unsigned int posDIN  = 3;	// output bit to configuration data input of XS board

// Set the value on the configuration initiation pin of the XC4000 FPGA
void XC40::SetPROGRAM(unsigned int b)
{
	SetControl(b,posPROG,posPROG);
}

// Get the value output on the configuration initiation pin
unsigned int XC40::GetPROGRAM(void)
{
	return GetControl(posPROG,posPROG);
}

// Toggle the configuration initiation pin (return it to its original level)
void XC40::PulsePROGRAM(void)
{
	assert(GetPROGRAM()==1);	// quiescent state of /PROGRAM should be one
	SetPROGRAM(~GetPROGRAM());	// toggle /PROGRAM output
	SetPROGRAM(~GetPROGRAM());	// toggle it again
}

// Set the value of the configuration clock bit
void XC40::SetCCLK(unsigned int b)
{
	SetControl(b,posCCLK,posCCLK);
}

// Get the value output on the configuration clock bit
unsigned int XC40::GetCCLK(void)
{
	return GetControl(posCCLK,posCCLK);
}

// Toggle the configuration clock (return it to its original level)
void XC40::PulseCCLK(void)
{
	assert(GetCCLK()==0);	// quiescent state of configuration clock should be zero
	SetCCLK(~GetCCLK());	// toggle clock output
	SetCCLK(~GetCCLK());	// toggle it again
}

// Initialize the XC4000 FPGA for configuration with DIN, CCLK, and PROGRAM pins
void XC40::InitConfigureFPLD(void)
{
	PulsePROGRAM();	// pulse the PROGRAM pin to start the configuration of the FPGA
	
	// insert 300 ms delay after PROGRAM pin goes high
	InsertDelay(300,MILLISECONDS);
}

// Set the value of the configuration data bit
void XC40::SetDIN(unsigned int b)
{
	SetControl(b,posDIN,posDIN);
}

// Get the value output on the configuration data bit
unsigned int XC40::GetDIN(void)
{
	return GetControl(posDIN,posDIN);
}

// Send out a byte of configuration information
void XC40::ConfigureFPLD(unsigned char b)
{
	// configuration bytes for the XC4000 programming pins only contain
	// values for the DIN pin, so eight clock pulses are needed per byte.
	for(int i=0; i<8; i++)
	{
		SetDIN(b&0x80 ? 1:0);  // byte is sent MSB first
		b <<= 1;
		PulseCCLK();
	}
}


// compute an integer from a vector of hex numbers in a stream
static long unsigned int GetInteger(istream& is, int len)
{
	assert(len<=4); // can't handle lengths greater than 32 bits
	long unsigned int lui = 0;
	for(int i=0; i<len; i++)
	{ // process hex digits starting from MSDigit
		unsigned char digit;
		is.read(&digit,1);
		assert(is.eof()==0);
		lui = lui*256 + (long unsigned int)digit;
	}
	return lui;
}

// get the field type indicator from a stream
static int GetType(istream& is)
{
	unsigned char type;
	is.read(&type,1);
	assert(is.eof()==0);
	return type;
}

// pass over fields until a certain field is found
static bool ScanForField(istream& is, unsigned char searchType)
{
	while(true)
	{
		unsigned char type = GetType(is);
		if(type==searchType)
			return true;
		unsigned int length = GetInteger(is);
		is.ignore(length);
		if(is.eof()!=0)
			break;
	}
	return false;
}

const int BitstreamFieldType = 0x65;	// field type for the bitstream data

// process a stream and send the configuration bitstream to the board
bool XC40::ConfigureFPLD(istream& is)
{
  unsigned long i;
	XSError& err = GetErr();
	
	if(is.eof()!=0)
		return false; // exit if no BIT stream exists
	
	// jump over the first field of the BIT file
	long unsigned int fieldLength = GetInteger(is);
	assert(fieldLength!=0);
	is.ignore(fieldLength);
	
	// process the second field
	fieldLength = GetInteger(is);
	assert(fieldLength==1);
	
	// now look for the field with the bitstream data
	bool status = ScanForField(is,BitstreamFieldType);
	assert(status==true);
	
	// now we are at the start of the configuration bits
	fieldLength = GetInteger(is,4); // get the length of the bit stream
	assert(fieldLength>0);
#if DEBUG
	cerr << "bit stream field length = " << fieldLength << "\n";
#endif
	InitConfigureFPLD(); // start the configuration process
	for(i=0; i<fieldLength; i++)
	{ // read the bytes and send the config. bits to the board
		unsigned char byte;
		is.read(&byte,1);
		assert(is.eof()==0);	// should not hit end-of-file
		ConfigureFPLD(byte);
		// output some feedback as the configuration proceeds
		if((i%1000)==0)
			cerr << "*";
	}
	cerr << "\n";
	cerr.flush();
	
	// clock a few more times at the end of the configuration process
	// to clean things up...
	for(i=0; i<4; i++)
		PulseCCLK();

	// insert a delay to let the chip initialize
	InsertDelay(100,MILLISECONDS);
	
#if DEBUG
	cerr << "exiting ConfigureFPLD()\n";
#endif
	
	return true;
}

const int DeviceFieldType = 0x62;		// field type for the FPGA device identifier

// gets the chip identifier from the bitstream and stores it in the object
bool XC40::SetChipType(istream& is)
{
  unsigned int i;
        XSError& err = GetErr();
	
	if(is.eof()!=0)
		return false; // exit if no BIT stream exists
	
	// jump over the first field of the BIT file
	long unsigned int fieldLength = GetInteger(is);
	assert(fieldLength!=0);
	is.ignore(fieldLength);
	
	// process the second field
	fieldLength = GetInteger(is);
	assert(fieldLength==1);
	
	// now look for the field with the chip identifier
	bool status = ScanForField(is,DeviceFieldType);
	assert(status==true);
	fieldLength = GetInteger(is);
	assert(fieldLength>0);
	char buf[8];
	for(i=0; i<7 && i<fieldLength; i++)
		is >> buf[i];
	buf[i] = 0;		// terminate string
	chipType = buf;	// store the chip identifier
	return true;	// everything went OK
}

// gets the chip identifier from the bitstream and stores it in the object
bool XC40::SetChipType(string& bitfileName)
{
	XSError& err = GetErr();
	
	if(bitfileName.compare("")==0)
		return false; // exit if no BIT file was given
	
	ifstream is(bitfileName.data(),ios::binary|ios::nocreate);
	if(is.fail()!=0)
	{ // error if given file could not be opened
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << bitfileName.data() << "\n";
		err.EndMsg();
	}
	
	// now get the chip identifier from the bitstream file
	bool status = SetChipType(is);
	
	is.close();
	
	return status;
}

// configures an FPGA so the XS Board RAM can be uploaded/downloaded
bool XC40::ConfigureRAMInterface()
{
	XSError& err = GetErr();
	
	string intfcCircuitFileName = ramInterfaceDir;
	intfcCircuitFileName += "\\D";
	intfcCircuitFileName += chipType;
	intfcCircuitFileName += ".BIT";
	return ramInterfaceIsRdy = XCPort::ConfigureFPLD(intfcCircuitFileName);
}

// download the XS Board RAM with the contents arriving through a stream
bool XC40::DownloadRAM(istream& is)
{
	XSError& err = GetErr();

	if(ramInterfaceIsRdy==false)
		if(ConfigureRAMInterface()==false)
			return false;

	// read hex records from file and place data in the RAM of the board
	while(is.eof()==0)
	{
		HexRecord hx;
		is >> hx;
		if(is.eof()!=0)
			break; // terminate loading loop when stream goes dry
		if(is.fail() || err.IsError())
		{
			err.SetSeverity(XSErrorFatal);
			err << "error reading stream\n";
			err.EndMsg();
			break;
		}
		if(hx.IsError())
		{ // some error in the hex record itself
			err.SetSeverity(XSErrorMajor);
			err << "hex record: " << hx.GetErrMsg() << "\n";
			err.EndMsg();
			break;
		}
		else
		{ // send out an indication for each hex record that's loaded
			cerr << ".";
			DownloadHexRecordToRAM(hx);
		}
	}
	cerr << "\n";
	cerr.flush();
	
	return true;
}


// bit pos of reset pin in par. port data reg for downloading circuit
static const unsigned int posRESET   = 0;
// bit pos of clock pin in par. port data reg for downloading circuit
static const unsigned int posCLK     = 1;
// bit pos of least-sig data-out pin in par. port data reg for downloading circuit
static const unsigned int posDOLSB   = 2;
// bit pos of most-sig data-out pin in par. port data reg for downloading circuit
static const unsigned int posDOMSB   = 5;
// bit pos of EEPROM output-enable control pin in par. port data reg for downloading circuit
static const unsigned int posEEOE = 6;
// bit pos of EEPROM chip-enable control pin in par. port data reg for downloading circuit
static const unsigned int posEECE    = 7;

// load RAM data at the addresses given in a hex record
void XC40::DownloadHexRecordToRAM(HexRecord& hx)
{
	if(hx.IsData()) // don't download unless this is RAM data
	{
		SetData(1,posRESET,posRESET);	// reset the downloading state machine
		SetData(1,posCLK,posCLK);		// force clock high
		SetData(0,posRESET,posRESET);	// release the reset

		unsigned int address = hx.GetAddress();	// get starting address for hex record
		SetData((address>>16)&0xf,posDOLSB,posDOMSB);
		SetData(0,posCLK,posCLK);	// latch address bits A19-A16 into downloading circuit
		SetData(1,posCLK,posCLK);
		SetData((address>>12)&0xf,posDOLSB,posDOMSB);
		SetData(0,posCLK,posCLK);	// latch address bits A15-A12 into downloading circuit
		SetData(1,posCLK,posCLK);
		SetData((address>> 8)&0xf,posDOLSB,posDOMSB);
		SetData(0,posCLK,posCLK);	// latch address bits A11-A8 into downloading circuit
		SetData(1,posCLK,posCLK);
		SetData((address>> 4)&0xf,posDOLSB,posDOMSB);
		SetData(0,posCLK,posCLK);	// latch address bits A7-A4 into downloading circuit
		SetData(1,posCLK,posCLK);
		SetData( address     &0xf,posDOLSB,posDOMSB);
		SetData(0,posCLK,posCLK);	// latch address bits A3-A0 into downloading circuit
		SetData(1,posCLK,posCLK);
	
		// process the bytes from a data-type hex record
		for(unsigned int i=0; i<hx.GetLength(); i++)
		{
//			int status = GetStatus(3,6);
//			if(status != 5)
//				fprintf(stdout,"incorrect download state %d (should be 5)\n",status);
			unsigned int data = hx[i];	// get next data byte from hex record
			SetData((data>>4)&0xf,posDOLSB,posDOMSB);
			SetData(0,posCLK,posCLK);	// latch data bits A7-A4 into downloading circuit
			SetData(1,posCLK,posCLK);
//			status = GetStatus(3,6);
//			if(status != 6)
//				fprintf(stdout,"incorrect download state %d (should be 6)\n",status);
			SetData( data    &0xf,posDOLSB,posDOMSB);
			SetData(0,posCLK,posCLK);	// send data bits A3-A0 through downloading circuit
			SetData(1,posCLK,posCLK);	// data is now loaded into RAM
		}

		SetData(1,posRESET,posRESET);	// reset the downloading state machine
		SetData(0,posRESET,posRESET);	// release the reset
	}
}


// upload the XS Board RAM to a stream
bool XC40::UploadRAM(
					   ostream& os,				// dump uploaded data to this stream
					   unsigned long loAddr,	// start fetching data from this address
					   unsigned long hiAddr		// stop at this address
					   )
{
	XSError& err = GetErr();

	if(ramInterfaceIsRdy==false)
		if(ConfigureRAMInterface()==false)
			return false;
	
	// read hex records from RAM and place data in the hex file
	for(unsigned long addr=loAddr&~0xF; addr<hiAddr; addr+=16)
	{
		HexRecord hx;
		UploadHexRecordFromRAM(hx,addr,addr+16); // get 16 bytes from RAM
		os << hx;		// send hex record to output stream
		cerr << ".";	// give feedback on progress
	}
	
	return true;
}

// bit pos of least-sig data-out pin in par. port data reg for downloading circuit
static const unsigned int posDILSB = 3;
// bit pos of most-sig data-out pin in par. port data reg for downloading circuit
static const unsigned int posDIMSB = 6;

// get RAM data at the addresses given and store it in a hex record
void XC40::UploadHexRecordFromRAM(
						HexRecord& hx,			// hex record to store data in
						unsigned long loAddr,	// begin upload from this address
						unsigned long hiAddr	// end upload at this address
						)
{
	XSError& err = GetErr();

	hx.SetTag(DataRecord);	// setup hex record to hold data
	hx.SetAddress(loAddr);	// set beginning and ending addresses for the hex record
	hx.SetLength(hiAddr-loAddr);
	
	SetData(1,posRESET,posRESET);	// reset the downloading state machine
	SetData(1,posCLK,posCLK);		// force clock high
	SetData(0,posRESET,posRESET);	// release the reset

	// set initial address into the downloading state machine
	SetData((loAddr>>16)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A19-A16 into downloading circuit
	SetData(1,posCLK,posCLK);
	SetData((loAddr>>12)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A15-A12 into downloading circuit
	SetData(1,posCLK,posCLK);
	SetData((loAddr>> 8)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A11-A8 into downloading circuit
	SetData(1,posCLK,posCLK);
	SetData((loAddr>> 4)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A7-A4 into downloading circuit
	SetData(1,posCLK,posCLK);
	SetData( loAddr     &0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A3-A0 into downloading circuit
	SetData(1,posCLK,posCLK);
	SetData(1,posRESET,posRESET);	// reset the downloading state machine
	SetData(0,posRESET,posRESET);	// release the reset

	int data;	// holds data read from RAM

	// process the bytes from a data-type hex record
	for(unsigned long i=loAddr+1; i<hiAddr; i++)
	{
		SetData((i>>16)&0xf,posDOLSB,posDOMSB);
		SetData(0,posCLK,posCLK);	// latch address bits A19-A16 into downloading circuit
		SetData(1,posCLK,posCLK);
		// read upper nybble of RAM data from previously loaded address
		data = GetStatus(posDILSB,posDIMSB);	
		SetData((i>>12)&0xf,posDOLSB,posDOMSB);
		SetData(0,posCLK,posCLK);	// latch address bits A15-A12 into downloading circuit
		SetData(1,posCLK,posCLK);
		// read lower nybble of RAM data from previously loaded address
		data = (data<<4) | (GetStatus(posDILSB,posDIMSB) & 0xf);
		hx[i-loAddr-1] = data;
		SetData((i>> 8)&0xf,posDOLSB,posDOMSB);
		SetData(0,posCLK,posCLK);	// latch address bits A11-A8 into downloading circuit
		SetData(1,posCLK,posCLK);
		SetData((i>> 4)&0xf,posDOLSB,posDOMSB);
		SetData(0,posCLK,posCLK);	// latch address bits A7-A4 into downloading circuit
		SetData(1,posCLK,posCLK);
		SetData( i     &0xf,posDOLSB,posDOMSB);
		SetData(0,posCLK,posCLK);	// latch address bits A3-A0 into downloading circuit
		SetData(1,posCLK,posCLK);
		SetData(1,posRESET,posRESET);	// reset the downloading state machine
		SetData(0,posRESET,posRESET);	// release the reset
	}

	// read the data from the last address in the range
	// load a dummy address just to get to the state where we can read next nybble of RAM
	SetData(((hiAddr-1)>>16)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);
	SetData(1,posCLK,posCLK);
	// read upper nybble of RAM data from last address
	data = GetStatus(posDILSB,posDIMSB);	
	SetData(((hiAddr-1)>>12)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);
	SetData(1,posCLK,posCLK);
	// read lower nybble of RAM data from last address
	data = (data<<4) | (GetStatus(posDILSB,posDIMSB) & 0xf);
	hx[hiAddr-loAddr-1] = data;
	SetData(1,posRESET,posRESET);	// reset the downloading state machine
	SetData(0,posRESET,posRESET);	// release the reset
	
	hx.CalcCheckSum();	// put the checksum into the hex record
}

// Send a byte into an Atmel EEPROM
void XC40::SendEEPROMByte(unsigned char byte)
{
	assert(GetCCLK()==0); // configuration clock should be zero on entry
	for(int i=0; i<8; i++)
	{
		// configuration data pin of parallel port connects to data pin of Atmel EEPROM
		SetDIN(byte&0x80 ? 1:0);  // byte is sent MSB first
		byte <<= 1;
		// CCLK pin of parallel port connects to clock pin of Atmel EEPROM
		PulseCCLK();
	}
	SetDIN(1);	// this is for the acknowledge bit from the EEPROM
	PulseCCLK();
	assert(GetCCLK()==0); // configuration clock should be zero on exit
}

// Program a page of an Atmel EEPROM
void XC40::ProgramEEPROMPage(unsigned char* buf,	// buffer with data bytes
							 unsigned int pageAddr,	// address where data will be stored
							 unsigned int len		// number of bytes in buffer
							 )
{
	assert(len<=64);	// an Atmel page is 64 bytes or less
	assert(pageAddr<0x20000L);	// Atmel byte addresses are less than 128K
	
	// initialize clock and data pins
	SetCCLK(0);
	SetDIN(1);
	// give start signal (data falls while clock is high)
	SetCCLK(1);
	SetDIN(0);
	SetCCLK(0);
	// send device-address byte with A2=0 and R/W=0 (write enabled)
	SendEEPROMByte(0xA6);
	// send two-byte address of EEPROM page
	SendEEPROMByte(pageAddr>>8);	// send upper byte
	SendEEPROMByte(pageAddr&0xFF);	// send lower bytes
	// send data bytes in page
	for(unsigned int i=0; i<len; i++)
		SendEEPROMByte(buf[i]);
	// give stop signal (data rises while clock is high)
	SetDIN(0);
	SetCCLK(1);
	SetDIN(1);
	// wait 10 ms
	InsertDelay(10,MILLISECONDS);
}

// program an Atmel serial EEPROM with a bitstream
bool XC40::ProgramEEPROM(istream& is)
{
  unsigned long i;
        XSError& err = GetErr();

	if(is.eof()!=0)
		return false; // exit if no BIT stream exists

	// We need to load the RAM interface circuit that is used to load the RAM
	// because it allows us to access the pins that are used to program
	// the serial EEPROM on the XS40 Board.
	if(ConfigureRAMInterface()==false)
		return false;

	// hold downloading state machine in the reset state so it won't move around
	// while the EEPROM is programmed
	SetData(1,posRESET,posRESET);	// reset the downloading state machine

	// Setup the pins of the Atmel EEPROM for programming.
	// The SER_EN_ should already be held low by a shunt on Jumper J6.
	// This holds the EEPROM CE_ low.
	SetData(0,posEECE,posEECE);
	SetData(0,posEEOE,posEEOE);
	
	// jump over the first field of the BIT file
	long unsigned int fieldLength = GetInteger(is);
#if DEBUG
	cerr << "first field length = " << fieldLength << "\n";
#endif
	assert(fieldLength!=0);
	is.ignore(fieldLength);
	
	// process the second field
	fieldLength = GetInteger(is);
#if DEBUG
	cerr << "second field length = " << fieldLength << "\n";
#endif
	assert(fieldLength==1);
	
	// now look for the field with the bitstream data
	bool status = ScanForField(is,BitstreamFieldType);
	assert(status==true);
	
	// now we are at the start of the configuration bits
	fieldLength = GetInteger(is,4); // get the length of the bit stream
	assert(fieldLength>0);
#if DEBUG
	cerr << "bit stream field length = " << fieldLength << "\n";
#endif
	
	// now load the EEPROM memory
	unsigned char page[64];
	for(i=0; i<fieldLength; i++)
	{ 
		is.read(&page[i&0x3f],1);	// read the bytes and into a 64-byte page
		if(is.eof())
		{
			break;
		}
		assert(is.eof()==0);	// shouldn't reach end-of-file
		if((i%1000)==0)		// after every 1000 bytes,
			cerr << "@";	// give some feedback during the programming process
		if((i&0x3f)==0x3f)
		{ // got a full 64-byte page, so program the EEPROM
			ProgramEEPROMPage(page,i&~0x3F,64);
		}
	}
	if((i&0x3F)!=0)
	{ // BIT file ended before a full 64-byte page was assembled
		ProgramEEPROMPage(page,i&~0x3F,(i&0x3F)+1);
	}

	// now set the output-enable polarity for the EEPROM to active-high
	SetData(1,posEECE,posEECE);	// raise chip-enable high
	SetData(1,posEEOE,posEEOE);	// raise output-enable high
	// write 0xFF to address 0x3FF to set the output-enable polarity
	unsigned char allOne = 0xFF;
	ProgramEEPROMPage(&allOne,0x3FFF,1);
		
	// programming done, so disable the EEPROM
	SetData(1,posEECE,posEECE);

	cerr << "\n";	// end the string of @ symbols
	cerr.flush();
	
	return true;
}
