/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The XC40 object specializes the XCPORT object so that the XC4000 FPGA on the
// XS40 Board can be configured and the RAM can be loaded.  The XC40 object controls
// the /PROGRAM, DIN, and CCLK pins.
//

// $Log: XC40.H,v $
// Revision 1.2  2000/11/14 11:25:16  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.4  1999/09/24 18:04:24  DEVB
// added routine to upload XS40 RAM to PC.
// changed download routines so they no longer use the pseudo-JTAG interface.
//
// Revision 1.3  1999/07/06 09:16:01  DEVB
// changed subroutine and variable names to reflect the fact that the
// FPGA must be configured for either uploading or downloading the RAM.
//
// Revision 1.2  1999/03/30 18:27:41  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:57:56  DEVB
// Initial revision
//

#ifndef XC40_H
#define XC40_H

#include "XCPORT.H"

class XC40 : public XCPort
{
	public:

	/* begin XC40 methods */

	/* method: XC40 */
	// Create an XC40 object.  The type of XC4000 FPGA on the XS40 Board is inferred
	// by reading the chip type from a bitfile that is passed to the constructor.
	// The directory containing the RAM interface bitstreams for all the different types
	// of XC4000 chips is also passed as an argument.
	 XC40(XSError* e, // error reporting channel
		unsigned int portNum, // parallel port number
		string& bitfileName, // bitfile with XC4000 configuration bitstream
		string& intfcDir // directory where RAM interface bitstreams are stored
		);

	/* method: ConfigureFPLD */
	// Send out a byte of configuration information
	void ConfigureFPLD(unsigned char b);

	/* method: ConfigureRAMInterface */
	// configures an FPGA so the XS Board RAM can be uploaded/downloaded
	bool ConfigureRAMInterface();

	/* method: ProgramEEPROM */
	// program an Atmel serial EEPROM with a bitstream
	bool ProgramEEPROM(istream& is);

	/* method: DownloadRAM */
	// download the XS Board RAM with the contents arriving through a stream
	bool DownloadRAM(istream& is);

	/* method: UploadRAM */
	// upload the XS Board RAM to a stream
	bool UploadRAM(
		ostream& os, // dump uploaded data to this stream
		unsigned long loAddr, // start fetching data from this address
		unsigned long hiAddr // stop at this address
		);

	private:

	/* method: DownloadHexRecordToRAM */
	// load RAM data at the addresses given in a hex record
	void DownloadHexRecordToRAM(HexRecord& hx);

	/* method: UploadHexRecordFromRAM */
	// get RAM data at the addresses given and store it in a hex record
	void UploadHexRecordFromRAM(
		HexRecord& hx, // hex record to store data in
		unsigned long loAddr, // begin upload from this address
		unsigned long hiAddr // end upload at this address
		);

	/* method: SetPROGRAM */
	// Set the value on the /PROGRAM pin of the XC4000 FPGA
	void SetPROGRAM(unsigned int b);

	/* method: GetPROGRAM */
	// Get the value output on the configuration initiation pin
	unsigned int GetPROGRAM(void);

	/* method: PulsePROGRAM */
	// Toggle the /PROGRAM output (return it to its original level)
	void PulsePROGRAM(void);

	/* method: SetCCLK */
	// Set the value of the configuration clock bit
	void SetCCLK(unsigned int b);

	/* method: GetCCLK */
	// Get the value output on the configuration clock bit
	unsigned int GetCCLK(void);

	/* method: PulseCCLK */
	// Toggle the CCLK output (return it to its original level)
	void PulseCCLK(void);

	/* method: SetDIN */
	// Set the value of the configuration data bit
	void SetDIN(unsigned int b);

	/* method: SetTMS */
	// Get the value output on the configuration data bit
	unsigned int GetDIN(void);

	/* method: InitConfigureFPLD */
	// Initialize the XC4000 FPGA for configuration with DIN, CCLK, and PROGRAM pins
	void InitConfigureFPLD(void);

	/* method: ConfigureFPLD */
	// field type for the bitstream data
	// process a stream and send the configuration bitstream to the board
	bool ConfigureFPLD(istream& is);

	/* method: SetChipType */
	// field type for the FPGA device identifier
	// gets the chip identifier from the bitstream and stores it in the object
	bool SetChipType(istream& is);

	/* method: SendEEPROMByte */
	// Send a byte into an Atmel EEPROM
	void SendEEPROMByte(unsigned char byte);

	/* method: ProgramEEPROMPage */
	// Program a page of an Atmel EEPROM
	void ProgramEEPROMPage(unsigned char* buf, // buffer with data bytes
		unsigned int pageAddr, // address where data will be stored
		unsigned int len // number of bytes in buffer
		);

	/* method: SetChipType */
	// gets the chip identifier from the bitstream and stores it in the object
	bool SetChipType(string& bitfileName);

	/* end XC40 methods */

	string ramInterfaceDir;	// directory where RAM interface bitstreams are found
};

/* begin XC40 functions */

/* function: GetInteger */
// compute an integer from a vector of hex numbers in a stream
static long unsigned int GetInteger(istream& is, int len=2);

/* function: GetType */
// get the field type indicator from a stream
static int GetType(istream& is);

/* function: ScanForField */
// pass over fields until a certain field is found
static bool ScanForField(istream& is, unsigned char searchType);

/* end XC40 functions */

#endif
