/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The XC95 object specializes the XCPORT object so that the XC9500 CPLD on the
// XS95 Board can be configured and the RAM can be loaded.  The XC95 object controls
// the JTAG port of the XC9500 chip.
//

// $Log: XC95.H,v $
// Revision 1.2  2000/11/14 11:25:16  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.4  1999/09/24 18:06:59  DEVB
// move upload and download XS95 RAM routines into the XC95 class.
//
// Revision 1.3  1999/07/06 09:16:01  DEVB
// changed subroutine and variable names to reflect the fact that the
// FPGA must be configured for either uploading or downloading the RAM.
//
// Revision 1.2  1999/03/30 18:27:41  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:59:47  DEVB
// Initial revision
//

#ifndef XC95_H
#define XC95_H

#include "XCPORT.H"

// definitions of XC9500 CPLD types
enum{
	XC9572Type = 0,
	XC95108Type = 1,
	UnknownXC9500Type = 2,
};

class XC95 : public XCPort
{
	public:

	/* begin XC95 methods */

	/* method: XC95 */
	// instantiate an object for loading an XS95 board on the parallel port
	 XC95(XSError* err, unsigned int portNum);

	/* method: ConfigureRAMInterface */
	// configure the CPLD so it can interface to the XS Board RAM
	bool ConfigureRAMInterface();

	/* method: ProgramEEPROM */
	// program an Atmel serial EEPROM with a bitstream
	bool ProgramEEPROM(istream& is);

	/* method: ConfigureFPLD */
	// Send out a byte of configuration information to the XC9500 Flash
	void ConfigureFPLD(unsigned char b);

	/* method: DownloadRAM */
	// download the XS Board RAM with the contents arriving through a stream
	bool DownloadRAM(istream& is);

	/* method: UploadRAM */
	// upload the XS Board RAM to a stream
	bool UploadRAM(
		ostream& os, // dump uploaded data to this stream
		unsigned long loAddr, // start fetching data from this address
		unsigned long hiAddr // stop at this address
		);

	private:

	/* method: DownloadHexRecordToRAM */
	// load RAM data at the addresses given in a hex record
	// (This routine assumes the entering TAP state is Shift-DR.
	// It exits with the TAP state being Shift-DR.)
	void DownloadHexRecordToRAM(HexRecord& hx);

	/* method: UploadHexRecordFromRAM */
	// get RAM data at the addresses given and store it in a hex record
	// (This routine assumes the entering TAP state is Shift-DR.
	// It exits with the TAP state being Shift-DR.)
	void UploadHexRecordFromRAM(
		HexRecord& hx, // hex record to store data in
		unsigned long loAddr, // begin upload from this address
		unsigned long hiAddr // end upload at this address
		);

	/* method: InitConfigureFPLD */
	// Initialize the XC9500 CPLD for configuration through the JTAG interface
	void InitConfigureFPLD(void);

	/* method: ConfigureFPLD */
	// Process SVF and send results through JTAG configuration pins
	bool ConfigureFPLD(istream& is);

	/* method: DetectXC9500Type */
	// detect the presence and type of the XC9500 CPLD chip on the port
	bool DetectXC9500Type();

	/* end XC95 methods */

	private:
};

/* begin XC95 functions */

/* function: NextWord */
static string NextWord(string& s);

/* end XC95 functions */

#endif
