/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// XCPort objects specialize a JTAGPort to handle programming of
// XS40 and XS95 Boards.
//

// $Log: XCPORT.CPP,v $
// Revision 1.4  2000/12/11 15:29:16  bretzdl
// exo file download support added, asm inline removed
//
// Revision 1.3  2000/12/01 17:40:17  bretzdl
// new version
//
// Revision 1.2  2000/11/14 11:25:16  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.9  2000/09/10 13:37:20  DEVB
// Replaced endl with newline to avoid problems with logging
// of messages by the xserror object.
//
// Revision 1.8  2000/02/15 05:23:19  DEVB
// changes to detect and report the absence of an XC9500 chip at the end of the
// download cable.
//
// Revision 1.7  1999/09/24 18:08:25  DEVB
// removed the RAM upload and download routines from the xcport class.
// the xc95 and xc40 classes now have their own separate versions of these
// routines.
// the routine for setting the XS Board programmable oscillator was also
// changed to reflect the new pin assignments for the osc. interface.
//
// Revision 1.6  1999/07/06 09:14:14  DEVB
// changed subroutine and variable names to reflect fact that FPGA/CPLD
// must be configured for either uploading or downloading the XS Board RAM.
//
// Revision 1.5  1999/06/29 16:56:27  DEVB
// adjusted time-delay routine to try and get more accuracy.
//
// Revision 1.4  1999/06/22 18:56:42  DEVB
// added routines to set the frequency of the programmable oscillator.
//
// Revision 1.3  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.2  1999/03/30 15:29:10  DEVB
// Changed InsertDelay function so that it has less chance of overflowing.
// Also disabled optimization from removing the empty timing loops.
//
// Revision 1.1  1999/03/29 21:59:44  DEVB
// Initial revision
//

//#include <dos.h>
#include <stdlib.h>
#include <stdio.h>
#include <fstream.h>
#include <strstream.h>
#include <time.h>
#include <string.h>
#include <unistd.h>

#include "XCPORT.H"
#include "UTILS.H"

// XCPort controls the JTAG port of the XC4000 and XC9500 chips, and also
// the XC4000 chip via its /PROGRAM, DIN, and CCLK pins.  The download/JTAG
// pins for the XS95 and XS40 boards are connected to the control outputs
// of the printer port (write address offset = 2) and the status inputs
// of the printer port (read address offset = 1).  The inversion mask for
// the outputs is:
//		1  C0-  drives /PROGRAM pin of the XC4000 directly,
//		0  C1-  drives TCK and CCLK through an inverter on the board,
//		0  C2+  drives TMS directly,
//		1  C3-  drives TDI and DIN directly.
// The inversion mask for the inputs is:
//		X  S0	unused
//		X  S1	unused
//		X  S2	unused
//		0  S3+  reads P15 of microcontroller directly,
//		0  S4+  reads P14 of microcontroller directly,
//		0  S5+  reads P16 of microcontroller directly,
//		0  S6+	reads P31 of microcontroller directly,
//		0  S7-  reads TDO/DOUT through an inverter.

// begin methods and functions

// instantiate an XCPort object on a given parallel port
XCPort::XCPort(XSError* e, unsigned int portNum) : JTAGPort(e,portNum,0x03,0,0x09)
{
	chipType = "UNKNOWN";
	ramInterfaceIsRdy = false;	// interface between FPLD and RAM does not start in ready state
}

// sets the chip identifier in the object
bool XCPort::SetChipType(string& type)
{
	int pos = type.find_first_of("0123456789");
	if(pos == -1)
	{	// no numbers!? not a xilinx chip id!
		chipType = "UNKNOWN";
		return false;
	}
	// remove any beginning non-numeric characters from chip type
	chipType = type.substr(pos,type.length()-pos);	
	int len = chipType.length();
	len = len>7 ? 7:len;	// keep only the first 7 characters
	chipType = chipType.substr(0,len);
	ConvertToUpperCase(chipType);
	return true;
}


bool XCPort::UploadRAM(
					   string& hexfileName,		// dump uploaded data to this file
					   unsigned long loAddr,	// start fetching data from this address
					   unsigned long hiAddr		// stop at this address
					   )
{
	XSError& err = GetErr();
	
	if(hexfileName.length()==0)
		return false;  // stop if no hex file was given
	
	cerr << "Uploading hex records to " << hexfileName.data()
                << " from XS Board RAM\n";
	cerr.flush();
	
	ofstream os(hexfileName.data());  // otherwise open hex file
	if(os.fail() || os.eof()!=0)
	{ // error - couldn't open hex file
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << hexfileName.data() << "\n";
		err.EndMsg();
	}
	
	UploadRAM(os,loAddr,hiAddr);
	
	os.close();  // close-up the hex file
	
	return true;
}
// download the XS Board RAM with the contents of a HEX file
bool XCPort::DownloadRAM(string& hexfileName)
{
	XSError& err = GetErr();
	
	if(hexfileName.length()==0)
		return false;  // stop if no hex file was given
	
	cerr << "Downloading hex records from " << hexfileName.data()
		<< " into XS Board RAM\n";
	cerr.flush();
	
	ifstream is(hexfileName.data(),ios::nocreate);  // otherwise open hex file
	if(is.fail() || is.eof()!=0)
	{ // error - couldn't open hex file
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << hexfileName.data() << "\n";
		err.EndMsg();
	}
	
	bool status = DownloadRAM(is);
	
	is.close();  // close-up the hex file
	
	return status;
}

// configure the FPLD with the bit stream stored in a BIT or SVF file
bool XCPort::ConfigureFPLD(string& fileName)
{
	XSError& err = GetErr();
	
	if(fileName.length()==0)
		return false;  // stop if no SVF or BIT file was given
	
	cerr << "Configuring FPLD with bitstream from " << fileName.data() << "\n";
	cerr.flush();
	
	ifstream is(fileName.data(),ios::binary|ios::nocreate);  // otherwise open BIT or SVF file
	if(is.fail() || is.eof())
	{ // error - couldn't open file
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << fileName.data() << "\n";
		err.EndMsg();
	}
	
	bool status = ConfigureFPLD(is);
	
	is.close();  // close-up the hex file
	
	return status;
}

// program the Atmel serial EEPROM on the XS Board with a bitstream from a file
bool XCPort::ProgramEEPROM(string& bitfileName)
{
	XSError& err = GetErr();
	
	if(bitfileName.compare("")==0)
		return false; // exit if no BIT file was given
	
	ifstream is(bitfileName.data(),ios::binary|ios::nocreate);
	if(is.fail() || is.eof())
	{ // error if given file could not be opened
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << bitfileName.data() << "\n";
		err.EndMsg();
	}
	
	// now configure the FPGA from the bitfile stream
	bool status = ProgramEEPROM(is);
	
	is.close();
	
	return status;
}


// set the frequency of the DS1075 programmable oscillator

const unsigned int OscPinPos = 0;		// pin thru which osc. receives frequency data  
const unsigned int OscPinTri = 1;		// pin which controls osc. pin tristate
const unsigned int OscCmdLength = 8;	// length of oscillator commands 
const unsigned int OscDataLength = 9;	// length of oscillator data
const unsigned int OscCmdWriteDIV = 1;	// command for writing clock divisor 
const unsigned int OscCmdWriteMUX = 2;	// command for writing clock multiplexer 

const unsigned int OneDuration = 5L;	// duration of a logic 1 in the osc. data stream
const unsigned int ZeroDuration = 112L;	// duration of a logic 0
const unsigned int BitDuration = 150L;	// total bit duration

// reset the programmable oscillator
void XCPort::ResetOsc()
{
	SetData(0,OscPinPos,OscPinPos);	// lower data bit 0
	InsertDelay(1200L,MICROSECONDS);	// wait for DS1075 to see reset level
	SetData(1,OscPinPos,OscPinPos);	// now remove reset level
	InsertDelay(1200L,MICROSECONDS);	// wait for presence pulse from DS1075
}

// send a bit of data to the programmable oscillator.
void XCPort::SendOscBit(unsigned char b)
{
	if(b)	// send a logic 1
	{
		SetData(0,OscPinPos,OscPinPos);
		InsertDelay(OneDuration,MICROSECONDS);
		SetData(1,OscPinPos,OscPinPos);
		InsertDelay(BitDuration-OneDuration,MICROSECONDS);
	}
	else	// send a logic 0
	{
		SetData(0,OscPinPos,OscPinPos);
		InsertDelay(ZeroDuration,MICROSECONDS);
		SetData(1,OscPinPos,OscPinPos);
		InsertDelay(BitDuration-ZeroDuration,MICROSECONDS);
	}
}

// send a command to the programmable oscillator
void XCPort::IssueOscCmd(unsigned int cmd)
{
	for(int i=0; i<OscCmdLength; i++, cmd>>=1)
		SendOscBit(cmd&1);
}

// send a data word to the programmable oscillator
void XCPort::SendOscData(unsigned int data)
{
	for(int i=0; i<OscDataLength; i++, data>>=1)
		SendOscBit(data&1);
}

// set the DS1075 oscillator frequency
void XCPort::SetOscFrequency(int div, int extOscPresent)
{
	int div1 = div==1 ? 1:0;	// set divide-by-1 bit if clock is not divided down
	int prescaleOff = 1;	// turn off prescalar
	int prescaleByTwo = 0;
	if(extOscPresent==0)	// setup prescalar circuitry if DS1075 internal oscillator is used
	{
		if(div>1026)
		{
			prescaleOff = 0;	// turn on the prescalar
			prescaleByTwo = 0;	// enable divide-by-four by turning off the divide-by-two prescalar
			div /= 4;			// reduce divisor
		}
		else if(div>513)
		{
			prescaleOff = 0;	// turn on the prescalar
			prescaleByTwo = 1;	// turn on the divide-by-two prescalar
			div /= 2;
		}
	}
	div = div>513 ? 513:div;	// divisor saturates at 513 if external oscillator is used
	// The divisor must be adjusted to get the frequency right as follows:
	//		original	adjusted	resulting
	//		div			div			frequency
	//		----------------------------------
	//		1			 -1			100 MHz (no division)
	//		2			  0			 50 MHz (divide by 2)
	//		3			  1			 33 MHz (divide by 3)
	//		4			  2			 25 MHz (divide by 4)
	//		...			...			...
	//		511			509			
	//		512			510			
	//		513			511
	div -= 2;

	int powerUp = 1;	// keep the oscillator powered up
	int disable0 = 1;	// disable the OUT0 output of the DS1075 (it's not connected on the XS Boards)
	int mux = (disable0<<5) | (powerUp<<4) | (prescaleByTwo<<3) |(prescaleOff<<2) | (div1<<1) | extOscPresent;

	SetData(1,OscPinPos,OscPinPos);	// make sure the osc. data pin starts out high
	SetData(0,OscPinTri,OscPinTri);	// activate the driver for the osc. pin
	// program divisor register of the DS1075
	ResetOsc();
	IssueOscCmd(OscCmdWriteDIV);
	SendOscData(div);
	InsertDelay(240,MICROSECONDS);
	// program multiplexor register of the DS1075
	ResetOsc();
	IssueOscCmd(OscCmdWriteMUX);
	SendOscData(mux);
	InsertDelay(240,MICROSECONDS);
	SetData(1,OscPinTri,OscPinTri);	// de-activate the driver for the osc. pin
}




// subroutine for delaying by a given number of microseconds or milliseconds
void XCPort::InsertDelay(unsigned long d, unsigned int time_units)
{
	static long loopsPerMillisecond = -1;
	unsigned long divider,rest;
	unsigned long loops,i2,i;
	clock_t start, finish;

	if(loopsPerMillisecond < 0)
        	{ // determine the timing of the computer if we haven't already done so
		printf("start time test\n");
                start = clock();        // get starting time
		for(i = 100000000L; i>0L; i--);       // do a whole bunch of empty loops
		finish = clock();       // get ending time
                printf("time test finish\n");
		assert(finish>0);
                // compute the number of empty loops per millisecond
                loopsPerMillisecond = 100000000L/(finish/1000);
                assert(loopsPerMillisecond>1000);       // we need enough precision
		printf("loops_per_milsec:%u\n",loopsPerMillisecond);
#if DEBUG
                cerr << "finish = " << finish << "   start = " << start << "\n";
                cerr << "loops per ms = " << loopsPerMillisecond << "\n";
                cerr.flush();
#endif

	}
	if(time_units==MICROSECONDS && d>10000)
	{
		time_units = MILLISECONDS;
		d /= 1000;
	}
	if(time_units == MILLISECONDS)
	{	// doing millisecond timing here
		for(loops = (loopsPerMillisecond*d); loops>0; loops--);
        }
	else
	{	// doing microsecond timing here
		for(loops = (loopsPerMillisecond*d)/1000; loops>0; loops--);
	}
}


