/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// XCPort objects specialize a JTAGPort to handle programming of
// XS40 and XS95 Boards.
//

// $Log: XCPORT.H,v $
// Revision 1.2  2000/11/14 11:25:16  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.5  1999/09/24 18:08:25  DEVB
// removed the RAM upload and download routines from the xcport class.
// the xc95 and xc40 classes now have their own separate versions of these
// routines.
// the routine for setting the XS Board programmable oscillator was also
// changed to reflect the new pin assignments for the osc. interface.
//
// Revision 1.4  1999/07/06 09:16:01  DEVB
// changed subroutine and variable names to reflect the fact that the
// FPGA must be configured for either uploading or downloading the RAM.
//
// Revision 1.3  1999/06/22 18:57:08  DEVB
// added routines to set the frequency of the programmable oscillator.
//
// Revision 1.2  1999/03/30 18:27:41  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 22:50:24  DEVB
// Initial revision
//

#ifndef XCPort_H
#define XCPort_H

#include <fstream.h>

#include "XCBSDR.H"
#include "JTAGPORT.H"
#include "HEXRECRD.H"

const unsigned int MILLISECONDS = 1000;
const unsigned int MICROSECONDS = 1;

class XCPort : public JTAGPort
{
	public:

	/* begin XCPort methods */

	/* method: XCPort */
	// instantiate an XCPort object on a given parallel port
	 XCPort(XSError* e, unsigned int portNum);

	/* method: SetChipType */
	// sets the chip identifier in the object
	bool SetChipType(string& type);
	
	// configure the FPLD so it can interface to the XS Board RAM
	virtual bool ConfigureRAMInterface() = 0;

	/* method: DownloadRAM */
	// download the XS Board RAM with the contents of a HEX file
	bool DownloadRAM(string& hexfileName);

	/* method: DownloadRAM */
	// download the XS Board RAM with the contents arriving through a stream
	virtual bool DownloadRAM(istream& is) = 0;

	/* method: DownloadHexRecordToRAM */
	// load RAM data at the addresses given in a hex record
	// (This routine assumes the entering TAP state is Shift-DR.
	// It exits with the TAP state being Shift-DR.)
	virtual void DownloadHexRecordToRAM(HexRecord& hx) = 0;

	/* method: UploadRAM */
	bool UploadRAM(
		string& hexfileName, // dump uploaded data to this file
		unsigned long loAddr, // start fetching data from this address
		unsigned long hiAddr // stop at this address
		);

	/* method: UploadRAM */
	// upload the XS Board RAM to a stream
	virtual bool UploadRAM(
		ostream& os, // dump uploaded data to this stream
		unsigned long loAddr, // start fetching data from this address
		unsigned long hiAddr // stop at this address
		) = 0;

	/* method: UploadHexRecordFromRAM */
	// get RAM data at the addresses given and store it in a hex record
	// (This routine assumes the entering TAP state is Shift-DR.
	// It exits with the TAP state being Shift-DR.)
	virtual void UploadHexRecordFromRAM(
		HexRecord& hx, // hex record to store data in
		unsigned long loAddr, // begin upload from this address
		unsigned long hiAddr // end upload at this address
		) = 0;

	/* method: ConfigureFPLD */
	// configure the FPLD with the bit stream stored in a BIT or SVF file
	bool ConfigureFPLD(string& fileName);
	
	/* method: ConfigureFPLD */
	// configure the FPLD with the bit stream arriving through a stream
	virtual bool ConfigureFPLD(istream& is) = 0;

	/* method: ProgramEEPROM */
	// program the Atmel serial EEPROM on the XS Board with a bitstream from a file
	bool ProgramEEPROM(string& bitfileName);

	/* method: ProgramEEPROM */
	// program EEPROM on the XS Board
	virtual bool ProgramEEPROM(istream& is) = 0;

	/* method: SetOscFrequency */
	// set the DS1075 oscillator frequency
	void SetOscFrequency(int div, int extOscPresent=0);

	/* method: InsertDelay */
	// subroutine for delaying by a given number of microseconds
	void InsertDelay(unsigned long d, unsigned int time_units);

	
	protected:
	
	// The RAM of the XS board is controlled through a JTAG interface,
	string chipType;		// type of chip (e.g. 4005XLPC84)
	XCBSDR* bsdrPtr;		// JTAG data register useful for XS RAM access
	Bitstream* bsirPtr;		// JTAG instruction register
	bool ramInterfaceIsRdy;	// true if the interface to up/download to the XS board RAM is ready


	private:

	/* method: ResetOsc */
	// reset the programmable oscillator
	void ResetOsc();

	/* method: SendOscBit */
	// send a bit of data to the programmable oscillator.
	void SendOscBit(unsigned char b);

	/* method: IssueOscCmd */
	// send a command to the programmable oscillator
	void IssueOscCmd(unsigned int cmd);

	/* method: SendOscData */
	// send a data word to the programmable oscillator
	void SendOscData(unsigned int data);

	/* end XCPort methods */

};

/* begin XCPort functions */

/* end XCPort functions */

#endif
