/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The XSV95 object specializes the XC95 object so that it can write and
// read the Flash device on the XSV Board.
//

// $Log: XSV95.CPP,v $
// Revision 1.3  2000/12/11 15:29:16  bretzdl
// exo file download support added, asm inline removed
//
// Revision 1.2  2000/11/14 11:25:16  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.1  2000/09/10 15:57:00  DEVB
// Initial revision
//

#include "XSV95.H"
#include <stdio.h>

// begin methods and functions

// instantiate an object for loading an XSV Board
XSV95::XSV95(XSError* err, unsigned int portNum, string& intfcDir) : XC95(err,portNum)
{
	flashInterfaceIsRdy = false;
	flashInterfaceDir = intfcDir;
}

// configure the XC95108 on the XSV for programming the Flash RAM
bool XSV95::ConfigureFlashInterface(void)
{
	XSError& err = GetErr();
	
	string intfcCircuitFileName = flashInterfaceDir;
	intfcCircuitFileName += "flashprg.svf";
	return flashInterfaceIsRdy = XCPort::ConfigureFPLD(intfcCircuitFileName);
}

// configure the XC95108 on the XSV for programming the Virtex from the Flash RAM
bool XSV95::ConfigureDefaultFlashConfigInterface(void)
{
	XSError& err = GetErr();
	
	string cfgCircuitFileName = flashInterfaceDir;
	cfgCircuitFileName += "flashcfg.svf";
	flashInterfaceIsRdy = false; // XC95108 no longer configured for programming the Flash RAM
	return XCPort::ConfigureFPLD(cfgCircuitFileName);
}

// download the XSV Flash with the contents of an EXO file
bool XSV95::DownloadFlash(string& exofileName)
{
	XSError& err = GetErr();
	
	if(exofileName.length()==0)
		return false;  // stop if no exo file was given
	
	cerr << "Downloading exo records from " << exofileName.data()
		<< " into XSV Board Flash\n";
	cerr.flush();
	
	ifstream is(exofileName.data(),ios::nocreate);  // otherwise open hex file
	if(is.fail() || is.eof()!=0)
	{ // error - couldn't open hex file
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << exofileName.data() << "\n";
		err.EndMsg();
	}
	
	DownloadFlash(is);
	
	is.close();  // close-up the exo file
	
	return true;
}

static const unsigned int flashSize = (1<<21);
static const unsigned int flashBlockSize = (1<<16);
static const unsigned int numBlocks = flashSize/flashBlockSize;
static const unsigned int blockAddressMask = flashBlockSize-1;
static bool blockErased[numBlocks];

// download the XSV Flash with the contents arriving through a stream
bool XSV95::DownloadFlash(istream& is)
{
	XSError& err = GetErr();

	if(flashInterfaceIsRdy==false)
		if(ConfigureFlashInterface()==false)
			return false;

	// initialize the block erasure flags
	for(int i=0; i<numBlocks; i++)
		blockErased[i] = false;

	// read hex records from file and place data in the RAM of the board
	while(is.eof()==0)
	{
		SRecord hx;
		is >> hx;
		if(is.eof()!=0)
			break; // terminate loading loop when stream goes dry
		if(is.fail() || err.IsError())
		{
			err.SetSeverity(XSErrorFatal);
			err << "error reading stream\n";
			err.EndMsg();
			break;
		}
		if(hx.IsError())
		{ // some error in the hex record itself
			err.SetSeverity(XSErrorMajor);
			err << "hex record: " << hx.GetErrMsg() << "\n";
			err.EndMsg();
			break;
		}
		else
		{ // send out an indication for each hex record that's loaded
			cerr << ".";
			DownloadHexRecordToFlash(hx);
		}
	}
	cerr << "\n";
	cerr.flush();
	
	return true;
}

// load RAM data at the addresses given in a hex record
void XSV95::DownloadHexRecordToFlash(SRecord& hx)
{
	if(hx.IsData()) // don't download unless this is Flash data
	{
		int address = hx.GetAddress();	// get starting address for hex record
		int i;
		for(i=address+hx.GetLength()-1; i>=address; i-=flashBlockSize)
		{
			if(blockErased[i/flashBlockSize]==false)
			{
				EraseFlashBlock(i);
				blockErased[i/flashBlockSize] = true;
			}
		}
		for(i=0; i<hx.GetLength(); i++)
			ProgramFlash(address+i,hx[i]);
	}
}

// upload the XSV Flash to a file
bool XSV95::UploadFlash(
					   string& exofileName,		// dump uploaded data to this file
					   unsigned long loAddr,	// start fetching data from this address
					   unsigned long hiAddr		// stop at this address
					   )
{
	XSError& err = GetErr();
	
	if(exofileName.length()==0)
		return false;  // stop if no exo file was given
	
	cerr << "Uploading exo records to " << exofileName.data()
                << " from XSV Board Flash\n";
	
	ofstream os(exofileName.data());  // otherwise open exo file
	if(os.fail() || os.eof()!=0)
	{ // error - couldn't open exo file
		err.SetSeverity(XSErrorFatal);
		err << "could not open " << exofileName.data() << "\n";
		err.EndMsg();
	}
	
	UploadFlash(os,loAddr,hiAddr);
	
	os.close();  // close-up the exo file
	
	return true;
}

// upload the XSV Flash to a stream
bool XSV95::UploadFlash(
					   ostream& os,				// dump uploaded data to this stream
					   unsigned long loAddr,	// start fetching data from this address
					   unsigned long hiAddr		// stop at this address
					   )
{
	XSError& err = GetErr();

//	if(ramInterfaceIsRdy==false)
//		if(ConfigureRAMInterface()==false)
//			return false;
	
	// read hex records from RAM and place data in the hex file
	for(unsigned long addr=loAddr&~0xF; addr<hiAddr; addr+=16)
	{
		SRecord hx;
		UploadHexRecordFromFlash(hx,addr,addr+16); // get 16 bytes from Flash
		os << hx;		// send hex record to output stream
		cerr << ".";	// give feedback on progress
	}
	
	return true;
}

// get Flash data at the addresses given and store it in a hex record
void XSV95::UploadHexRecordFromFlash(
						SRecord& hx,			// hex record to store data in
						unsigned long loAddr,	// begin upload from this address
						unsigned long hiAddr	// end upload at this address
						)
{
	XSError& err = GetErr();

	hx.SetTag(DataSRecord);	// setup hex record to hold data
	hx.SetAddress(loAddr);	// set beginning and ending addresses for the hex record
	hx.SetLength(hiAddr-loAddr);
	
	ResetFlash();

	// process the bytes from a data-type hex record
	for(unsigned long i=loAddr; i<hiAddr; i++)
	{
		hx[i-loAddr] = ReadFlashByte(i);
	}
	
	hx.CalcCheckSum();	// put the checksum into the hex record
}


// bit pos of reset pin in par. port data reg for downloading circuit
static const unsigned int posRESET   = 0;
// bit pos of clock pin in par. port data reg for downloading circuit
static const unsigned int posCLK     = 1;
// bit pos of least-sig data-out pin in par. port data reg for downloading circuit
static const unsigned int posDOLSB   = 2;
// bit pos of most-sig data-out pin in par. port data reg for downloading circuit
static const unsigned int posDOMSB   = 5;
// bit pos of least-sig data-input pin in par. port status reg for downloading circuit
static const unsigned int posDILSB = 3;
// bit pos of most-sig data-input pin in par. port status reg for downloading circuit
static const unsigned int posDIMSB = 6;

// read a byte from the Flash through the parallel port
unsigned int XSV95::ReadFlashByte(unsigned int address)
{
	SetData(1,posRESET,posRESET);	// reset the downloading state machine
	SetData(1,posCLK,posCLK);		// force clock high
	SetData(0,posRESET,posRESET);	// release the reset

	int status = GetStatus(posDILSB,posDIMSB);
	if(status != 0)
		fprintf(stdout,"incorrect download state %d (should be 0)\n",status);

	SetData((address>>20)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A23-A20 into downloading circuit
	SetData(1,posCLK,posCLK);
	
	SetData((address>>16)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A19-A16 into downloading circuit
	SetData(1,posCLK,posCLK);
	
	SetData((address>>12)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A15-A12 into downloading circuit
	SetData(1,posCLK,posCLK);
	
	status = GetStatus(posDILSB,posDIMSB);
	if(status != 3)
		fprintf(stdout,"incorrect download state %d (should be 3)\n",status);
	
	SetData((address>> 8)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A11-A8 into downloading circuit
	SetData(1,posCLK,posCLK);
	
	status = GetStatus(posDILSB,posDIMSB);
	if(status != 4)
		fprintf(stdout,"incorrect download state %d (should be 4)\n",status);
	
	SetData((address>> 4)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A7-A4 into downloading circuit
	SetData(1,posCLK,posCLK);
	
	status = GetStatus(posDILSB,posDIMSB);
	if(status != 5)
		fprintf(stdout,"incorrect download state %d (should be 5)\n",status);
	
	SetData( address     &0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);		// latch address bits A3-A0 into downloading circuit
	SetData(1,posCLK,posCLK);
	SetData(0,posCLK,posCLK);

	SetData(1,posRESET,posRESET);	// reset the downloading state machine
	SetData(0,posRESET,posRESET);	// release the reset

	status = GetStatus(posDILSB,posDIMSB);
	if(status != 0)
		fprintf(stdout,"incorrect download state %d (should be 0)\n",status);

	SetData((address>>20)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A23-A20 into downloading circuit
	SetData(1,posCLK,posCLK);
	// read upper nybble of RAM data from previously loaded address
	unsigned int data = GetStatus(posDILSB,posDIMSB);	
	SetData((address>>16)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A19-A16 into downloading circuit
	SetData(1,posCLK,posCLK);
	// read lower nybble of RAM data from previously loaded address
	data = (data<<4) | (GetStatus(posDILSB,posDIMSB) & 0xf);
	return data;
}

// write a byte to the Flash through the parallel port
void XSV95::WriteFlashByte(unsigned int address, unsigned int data)
{
	SetData(1,posRESET,posRESET);	// reset the downloading state machine
	SetData(1,posCLK,posCLK);		// force clock high
	SetData(0,posRESET,posRESET);	// release the reset

	int status = GetStatus(posDILSB,posDIMSB);
	if(status != 0)
		fprintf(stdout,"incorrect download state %d (should be 0)\n",status);

	SetData((address>>20)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A23-A20 into downloading circuit
	SetData(1,posCLK,posCLK);
	
	SetData((address>>16)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A19-A16 into downloading circuit
	SetData(1,posCLK,posCLK);
	
	SetData((address>>12)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A15-A12 into downloading circuit
	SetData(1,posCLK,posCLK);
	
	status = GetStatus(posDILSB,posDIMSB);
	if(status != 3)
		fprintf(stdout,"incorrect download state %d (should be 3)\n",status);

	SetData((address>> 8)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A11-A8 into downloading circuit
	SetData(1,posCLK,posCLK);
	
	status = GetStatus(posDILSB,posDIMSB);
	if(status != 4)
		fprintf(stdout,"incorrect download state %d (should be 4)\n",status);
	
	SetData((address>> 4)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch address bits A7-A4 into downloading circuit
	SetData(1,posCLK,posCLK);
	
	status = GetStatus(posDILSB,posDIMSB);
	if(status != 5)
		fprintf(stdout,"incorrect download state %d (should be 5)\n",status);
	
	SetData( address     &0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);		// latch address bits A3-A0 into downloading circuit
	SetData(1,posCLK,posCLK);

	status = GetStatus(posDILSB,posDIMSB);
	if(status != 6)
		fprintf(stdout,"incorrect download state %d (should be 6)\n",status);

	SetData((data>>4)&0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch data bits D7-D4 into downloading circuit
	SetData(1,posCLK,posCLK);

	SetData( data    &0xf,posDOLSB,posDOMSB);
	SetData(0,posCLK,posCLK);	// latch data bits D3-D0 into downloading circuit
	SetData(1,posCLK,posCLK);
}

// reset the Flash so data can be read from it
void XSV95::ResetFlash()
{
	WriteFlashByte(0,0xFF);
}

// read data from the Flash
unsigned int XSV95::ReadFlashID()
{
	WriteFlashByte(0,0x90);
	return ReadFlashByte(0x0) | (ReadFlashByte(0x1)<<8);
}

// get the status of the Flash
unsigned int XSV95::ReadFlashStatus()
{
	WriteFlashByte(0,0x70);
	return ReadFlashByte(0);
}

// clear any error bits in the Flash status register
void XSV95::ClearFlashStatus()
{
	WriteFlashByte(0,0x50);
}

static const unsigned int rdyBitMask = 1<<7;

// lock a Flash block so it can't be altered
static const unsigned int lockBlockErrorMask = (1<<5)|(1<<4)|(1<<3)|(1<<1);
void XSV95::LockFlashBlock(unsigned int address)
{
	WriteFlashByte(address,0x60);
	WriteFlashByte(address,0x01);
	unsigned int status = ReadFlashStatus();
	while((status & rdyBitMask) == 0)
		status = ReadFlashStatus();
	if((status & lockBlockErrorMask) != 0)
	{
		fprintf(stdout,"error locking block @ %06x: %02x\n",address,status);
		ClearFlashStatus();
	}
}

// lock all the block lock bits so they cannot be altered
void XSV95::MasterLockFlash()
{
	WriteFlashByte(0,0x60);
	WriteFlashByte(0,0xF1);
	unsigned int status = ReadFlashStatus();
	while((status & rdyBitMask) == 0)
		status = ReadFlashStatus();
	if((status & lockBlockErrorMask) != 0)
	{
		fprintf(stdout,"error locking block lock bits: %02x\n",status);
		ClearFlashStatus();
	}
}

// unlock all the Flash blocks
static const unsigned int unlockBlockErrorMask = (1<<5)|(1<<4)|(1<<3)|(1<<1);
void XSV95::UnlockFlashBlocks()
{
	WriteFlashByte(0,0x60);
	WriteFlashByte(0,0xD0);
	unsigned int status = ReadFlashStatus();
	while((status & rdyBitMask) == 0)
		status = ReadFlashStatus();
	if((status & unlockBlockErrorMask) != 0)
	{
		fprintf(stdout,"error unlocking blocks: %02x\n",status);
		ClearFlashStatus();
	}
}


static const unsigned int programErrorMask = (1<<4)|(1<<3)|(1<<1);

// Program a byte in the Flash
void XSV95::ProgramFlash(unsigned int address, unsigned int data)
{
//	fprintf(stdout,"%02x => (%06x)\n",data,address);
	WriteFlashByte(address,0x40);
	WriteFlashByte(address,data);
	unsigned int status = ReadFlashStatus();
	while((status & rdyBitMask) == 0)
		status = ReadFlashStatus();
	if((status & programErrorMask) != 0)
	{
		fprintf(stdout,"error programming byte @ %06x: %02x\n",address,status);
		ClearFlashStatus();
	}
}


static const unsigned int eraseErrorMask = (1<<5)|(1<<4)|(1<<3)|(1<<1);

// erase the contents of a Flash block
void XSV95::EraseFlashBlock(unsigned int address)
{
	WriteFlashByte(address,0x20);
	WriteFlashByte(address,0xD0);
	unsigned int status = ReadFlashStatus();
	while((status & rdyBitMask) == 0)
		status = ReadFlashStatus();
	if((status & eraseErrorMask) != 0)
	{
		fprintf(stdout,"error erasing @ %06x: %02x\n",address,status);
		ClearFlashStatus();
	}
}

