/*
 * This software is in the public domain.
 * You may use it as you see fit with the understanding that
 * we make no guarantees of its fitness for any particular purpose.
 *
 * Author: XESS Corporation
 *         www.xess.com
 */

//
// The XSLOAD program has the following functions:
//		1. Programs XS40 Boards with XC4000 BIT files
//		2. Programs XS95 Boards with XC9500 SVF files
//		3. Loads the XS Board RAM with HEX data files
//		4. Programs the Serial EEPROM on the XS40 Board
//

// $Log: XSLOAD.CPP,v $
// Revision 1.3  2000/12/11 15:29:16  bretzdl
// exo file download support added, asm inline removed
//
// Revision 1.2  2000/11/14 11:25:16  bretzdl
// Port to Linux. Makefile DLPORTIO.C added
//
// Revision 1.8  2000/09/10 17:51:48  DEVB
// Replaced endl with newline since xserror objects don't like endl anymore.
//
// Revision 1.7  1999/09/24 18:24:08  DEVB
// added options for selecting programmable oscillator clock source
// and for selecting the directory where the RAM interface bitstreams
// are stored.
//
// Revision 1.6  1999/07/06 09:18:52  DEVB
// added option to upload data from XS Board RAM to a hex file.
//
// Revision 1.5  1999/06/29 16:49:04  DEVB
// added ability to set programmable osc on XS95 Board.
//
// Revision 1.4  1999/06/23 02:11:18  DEVB
// added option to set clock divisor of the DS1075 programmable oscillator
// on the XS40 Board.
//
// Revision 1.3  1999/03/30 19:46:28  DEVB
// Added code to display release version number
//
// Revision 1.2  1999/03/30 18:27:12  DEVB
// Released source into the public domain
//
// Revision 1.1  1999/03/29 21:49:31  DEVB
// Initial revision
//

#include <stdlib.h>
#include <assert.h>
#include <strstream.h>
#include <fstream.h>
#include <string>
using std::string;

#include "../XC40.H"
#include "../XC95.H"
#include "../XSV95.H"
#include "../HEX.H"
#include "../UTILS.H"

//
// global variables
//
string version("$Name:  $");
XSError errMsg(cerr);		// setup error channel
unsigned int portNum = 1;	// default parallel port number
bool uploadRAM = false;		// enable reading of XS Board RAM
unsigned int loAddr, hiAddr;	// address boundaries for reading RAM
int clockDivisor = -1;		// divisor to be loaded into XS Board programmable oscillator
int extOscPresent = 0;		// flag indicates presence of external oscillator to DS1075
string loadFileName = "";	// name of 32 KByte RAM loading circuit
string hexFileName  = "";	// name of file to be loaded into RAM (.HEX)
string svfFileName  = "";	// name of run-time circuit (.SVF) for XC9500
string bitFileName  = "";	// name of run-time circuit (.BIT) for XC4000
string exoFileName  = "";
string dumpFileName = "";	// name of RAM data dump file
unsigned int numHexFiles     = 0;	// number of hex data files on the command line
unsigned int numSvfFiles     = 0;	// number of XC9500 SVF files on the command line
unsigned int numBitFiles     = 0;	// number of XC4000 BIT files on the command line
unsigned int numExoFiles     = 0;       // number of XC9500 EXO files on the command line
string xstoolsBinDir = "";	// directory containing this program executable
bool programSerialEEPROM = false;

//
// print a string that shows the options and how to use this program.
//
void Usage(const string& progName)
{
	cerr << progName.data() 
		<< "\n\t[-[h|help]]: get help"
		<< "\n\t[-[p|port] [1,2,3]]: select parallel port"
		<< "\n\t[-serial_eeprom]: program serial EEPROM on XS40 Board"
		<< "\n\t[-upload lo_address hi_address]: upload XS Board RAM contents between low and high addresses"
		<< "\n\t[-clock_divisor #]: program the divisor in the programmable oscillator"
		<< "\n\t[-external_clock]: enable the external clock input of the programmable oscillator"
		<< "\n\tfile1[.HEX]: HEX file that will be download/uploaded to/from XS Board RAM"
		<< "\n\tfile2[.BIT|.SVF]: bitstream file that will be downloaded to XS Board FPGA/CPLD"
		<< endl;
	int pos = version.find("Name:");
	if(pos < 0)
	{
		cerr << "Version number unavailable!!" << endl;
	}
	else
	{
		string versionNumber = version.substr(pos+5,version.length()-pos-5);
		for(int i=versionNumber.length()-1; i>=0; i--)
		{
			if(versionNumber[i]=='_')
				versionNumber[i] = '.';
			else if(versionNumber[i]=='$')
				versionNumber[i] = ' ';
		}
		cerr << "Version " << versionNumber.data() << endl;
	}
}

//
// process the command line options and set the appropriate flags.
//
void ProcessOpts(int* argc, char** argv)
{
        int i,j;
	char* dir = NULL;
	bool oldBoardType = false;

	for(i=1; i< *argc; i++)
	{
		string arg = argv[i];
		
		if(arg[0] == '-')  // found an option
		{
			int cnt, j;
			ConvertToUpperCase(arg);	// convert option to upper case
			
			// ask for help but we don't give much
			if(arg=="-H" || arg=="-HELP")
			{
				Usage(arg);
				cnt = 1;
			}
			
			// specify the printer port being used
			else if(arg=="-PORT" || arg=="-P")
			{
				istrstream argstr(argv[i+1]);
				argstr >> portNum;
				cnt = 2;
			}
			
			// specify that the serial EEPROM is to be programmed
			else if(arg=="-SERIAL_EEPROM" || arg=="-SEEP")
			{
				programSerialEEPROM = true;
				cnt = 1;
			}

			// specify what range of XS Board RAM addresses should be uploaded into a file
			else if(arg=="-UPLOAD" || arg=="-U")
			{
				uploadRAM = true;
				istrstream argstr1(argv[i+1]);
				argstr1 >> loAddr;
				istrstream argstr2(argv[i+2]);
				argstr2 >> hiAddr;
				cnt = 3;
			}
			
			// specify the clock divisor for the DS1075 programmable oscillator
			else if(arg=="-CLOCK_DIVISOR" || arg=="-DIV")
			{
				istrstream argstr(argv[i+1]);
				argstr >> clockDivisor;
				cnt = 2;
			}
			
			// specify the clock source for the DS1075 programmable oscillator
			else if(arg=="-EXTERNAL_CLOCK" || arg=="-EXTERNAL")
			{
				extOscPresent = 1;
				cnt = 1;
			}
			
			// specify the directory where the RAM interface bitstreams are stored
			else if(arg=="-DIR")
			{
				dir = argv[i+1];
				cnt = 2;
			}

			// specify whether board is pre-128K SRAM type
			else if(arg=="-OLD")
			{
				oldBoardType = true;
				cnt = 1;
			}
			
			// can't figure out what the user wants
			else
			{
				errMsg.SetSeverity(XSErrorMinor);
				errMsg << "unknown option: " << arg.data() << "\n";
				errMsg.EndMsg();
				Usage(string(argv[0]));
				cnt = 1;
			}
			
			// remove the processed options from the argument list
			for(j=i+cnt; j< *argc; j++)
				argv[j-cnt] = argv[j];
			*argc -= cnt;
			i--;
		}
	}
	
	// all the options are processed, now handle the download files from
	// the command line
	string suffixes[] = {"SVF","BIT","HEX","EXO"};
	int numFiles[4] = {0,};		// records number of each file type in command line
	string fileName[4];			// file names for SVF, BIT, and HEX files
	for(i=1; i<*argc; i++)
	{
		ifstream fs;
		string arg = argv[i];
		string suffix = GetSuffix(arg);
		ConvertToUpperCase(suffix);
		if(suffix.empty()==true)
		{ // no suffix, so see what kinds of files are present
			bool found = false;
			for(int j=0; j<4 && found==false; j++)
			{
				string tmpFileName = arg + "." + suffixes[j];
				fs.open(tmpFileName.data(),ios::nocreate);
				if(fs.good())
				{ // yes, a file with this suffix and base name exists
					found = true;
					numFiles[j]++;
					fileName[j] = tmpFileName;
					fs.close();
				}
			}
			if(found==false)
			{ // no file with any of the allowed suffixes was found
				errMsg.SetSeverity(XSErrorMinor);
				errMsg << "can't find .BIT, .SVF, or .HEX file for this: "
					<< arg.data() << "\n";
				errMsg.EndMsg();
				Usage(string(argv[0]));
				return;
			}
		}
		else
		{ // suffix found, so see what the file suffix is
			for(j=0; j<4; j++)
			{
				if(suffix == suffixes[j])
				{
					numFiles[j]++;
					fileName[j] = argv[i];
					break;
				}
			}
			if(j>=4)
			{
				errMsg.SetSeverity(XSErrorMinor);
				errMsg << "unknown file type: " << arg.data() << "\n";
				errMsg.EndMsg();
				Usage(string(argv[0]));
				return;
			}
		}
	}
	numSvfFiles += numFiles[0];
	numBitFiles += numFiles[1];
	numHexFiles += numFiles[2];
	numExoFiles += numFiles[3];
	svfFileName = fileName[0];
	bitFileName = fileName[1];
	hexFileName = fileName[2];
	exoFileName = fileName[3];
	
	if(numHexFiles>1)
	{
		errMsg.SetSeverity(XSErrorMinor);
		errMsg << "only one HEX file is allowed" << "\n";
		errMsg.EndMsg();
		Usage(string(argv[0]));
	}
	if(numBitFiles>1)
	{
		errMsg.SetSeverity(XSErrorMinor);
		errMsg << "only one BIT file is allowed" << "\n";
		errMsg.EndMsg();
		Usage(string(argv[0]));
	}
	if(numSvfFiles>1)
	{
		errMsg.SetSeverity(XSErrorMinor);
		errMsg << "only one SVF file is allowed" << "\n";
		errMsg.EndMsg();
		Usage(string(argv[0]));
	}
	if(numExoFiles>1)
	{
		errMsg.SetSeverity(XSErrorMinor);
		errMsg << "only one EXO file is allowed" << "\n";
		errMsg.EndMsg();
		Usage(string(argv[0]));
	}
	
	if(numBitFiles==0 && numSvfFiles==0  && numHexFiles==0 && numExoFiles==0)
	{
		errMsg.SetSeverity(XSErrorMinor);
		errMsg << "at least one BIT, SVF, or HEX file is needed" << "\n";
		errMsg.EndMsg();
		Usage(string(argv[0]));
	}

	if(dir==NULL)
	{
		dir = getenv("XSTOOLS_BIN_DIR");
		if(dir==NULL)
		{
			errMsg.SetSeverity(XSErrorFatal);
			errMsg << "XSTOOLS_BIN_DIR environment variable is not set" << "\n";
			errMsg.EndMsg();
		}
	}
	xstoolsBinDir = dir;
	if(oldBoardType)
		xstoolsBinDir = xstoolsBinDir + "\\old";
}


main( int argc, char **argv )
{
	// set the error message header to the name of this program
	string cmd = argv[0];
	cmd = StripSuffix(StripPrefix(cmd));
	errMsg.SetHeader(cmd);
	
	ProcessOpts( &argc, argv );  // process command line options
	if(errMsg.IsError()) // exit if errors in options
		exit(1);
	
	
	XSV95 xsv95Port(&errMsg,portNum,xstoolsBinDir);

	XCPort* portPtr;
	if(numBitFiles>0)
		portPtr = new XC40(&errMsg,portNum,bitFileName,xstoolsBinDir);
	else
		portPtr = new XC95(&errMsg,portNum);
	XCPort& port = *portPtr;
	
	if(clockDivisor>0)
	{
		if(numBitFiles>0)
		{
			port.ConfigureFPLD(bitFileName);
		}
		else if(numSvfFiles>0)
		{
			port.ConfigureFPLD(svfFileName);
		}
		port.SetOscFrequency(clockDivisor,extOscPresent);
		exit(0);
	}

	if(uploadRAM==true)
	{
		port.UploadRAM(hexFileName,loAddr,hiAddr);
		exit(0);
	}

	if(numHexFiles>0)
	{
		port.DownloadRAM(hexFileName);
		if(errMsg.IsError())
			exit(1);
	}
	
	if(numBitFiles>0)
	{
		if(programSerialEEPROM==true)
			port.ProgramEEPROM(bitFileName);
		else
			port.ConfigureFPLD(bitFileName);
		if(errMsg.IsError())
			exit(1);
	}
	else if(numSvfFiles>0)
	{
		port.ConfigureFPLD(svfFileName);
		if(errMsg.IsError())
			exit(1);
	}
	else if(numExoFiles>0)
	{
		xsv95Port.DownloadFlash(exoFileName);
		if(errMsg.IsError())
			exit(1);
	}
	exit(0);
}
